#!/usr/bin/env python
"""
Determine number of processors and memory for the worker node
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script
from DIRAC import gLogger
from DIRAC.WorkloadManagementSystem.Utilities import JobParameters

ceName = ""
ceType = ""
Queue = ""


def setCEName(args):
    global ceName
    ceName = args


def setSite(args):
    global Site
    Site = args


def setQueue(args):
    global Queue
    Queue = args


@Script()
def main():
    global ceName
    global Site
    global Queue
    Script.registerSwitch("N:", "Name=", "Computing Element Name (Mandatory)", setCEName)
    Script.registerSwitch("S:", "Site=", "Site Name (Mandatory)", setSite)
    Script.registerSwitch("Q:", "Queue=", "Queue Name (Mandatory)", setQueue)
    Script.parseCommandLine(ignoreErrors=True)

    gLogger.info("Getting number of processors")
    numberOfProcessor = JobParameters.getNumberOfProcessors(Site, ceName, Queue)

    gLogger.info("Getting memory (RAM) from MJF")
    maxRAM = JobParameters.getMemoryFromMJF()
    if not maxRAM:
        gLogger.info("maxRAM could not be found in MJF, using JobParameters.getMemoryFromProc()")
        maxRAM = JobParameters.getMemoryFromProc()

    gLogger.info("Getting number of GPUs")
    numberOfGPUs = JobParameters.getNumberOfGPUs(Site, ceName, Queue)

    # just communicating it back
    gLogger.notice(" ".join(str(wnPar) for wnPar in [numberOfProcessor, maxRAM, numberOfGPUs]))


if __name__ == "__main__":
    main()



"""
Wrapper on top of ElasticDB. It is used to manage the DIRAC monitoring types.

**Configuration Parameters**:

The following option can be set in `Systems/Monitoring/<Setup>/Databases/MonitoringDB`

* *IndexPrefix*:  Prefix used to prepend to indexes created in the ES instance. If this
                  is not present in the CS, the indexes are prefixed with the setup name.

For each monitoring types managed, the Period (how often a new index is created)
can be defined with::

   MonitoringTypes
   {
     ComponentMonitoring
     {
       # Indexing strategy. Possible values: day, week, month, year, null
       Period = month
     }
     RMSMonitoring
     {
       # Indexing strategy. Possible values: day, week, month, year, null
       Period = month
     }
     WMSHistory
     {
       # Indexing strategy. Possible values: day, week, month, year, null
       Period = day
     }
   }


"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import time
import calendar

from DIRAC import S_OK, S_ERROR
from DIRAC.Core.Base.ElasticDB import ElasticDB
from DIRAC.Core.Utilities.Plotting.TypeLoader import TypeLoader
from DIRAC.ConfigurationSystem.Client.Helpers import CSGlobals
from DIRAC.ConfigurationSystem.Client.Config import gConfig
from DIRAC.ConfigurationSystem.Client.PathFinder import getDatabaseSection


########################################################################
class MonitoringDB(ElasticDB):
    """Extension of ElasticDB for Monitoring system DB"""

    def __init__(self, name="Monitoring/MonitoringDB", readOnly=False):
        """Standard constructor"""

        try:
            section = getDatabaseSection("Monitoring/MonitoringDB")
            indexPrefix = gConfig.getValue("%s/IndexPrefix" % section, CSGlobals.getSetup()).lower()

            # Connecting to the ES cluster
            super(MonitoringDB, self).__init__(dbname=name.split("/")[1], fullName=name, indexPrefix=indexPrefix)
        except RuntimeError as ex:
            self.log.error("Can't connect to MonitoringDB", repr(ex))
            raise ex

        self.__readonly = readOnly
        self.documentTypes = {}

        # loads all monitoring indexes and types.
        objectsLoaded = TypeLoader("Monitoring").getTypes()

        # Load the files
        for pythonClassName in sorted(objectsLoaded):
            typeClass = objectsLoaded[pythonClassName]
            indexName = "%s_%s" % (self.getIndexPrefix(), typeClass()._getIndex())
            monitoringType = typeClass().__class__.__name__
            mapping = typeClass().mapping
            monfields = typeClass().monitoringFields
            period = self.getCSOption("MonitoringTypes/%s/Period" % monitoringType, typeClass().period)
            self.documentTypes[monitoringType] = {
                "indexName": indexName,
                "mapping": mapping,
                "monitoringFields": monfields,
                "period": period,
            }
            if self.__readonly:
                self.log.info("Read only mode: no new index will be created")
            else:
                # Verifying if the index is there, and if not create it
                res = self.existingIndex("%s-*" % indexName)  # check with a wildcard
                if res["OK"] and res["Value"]:
                    actualIndexName = self.generateFullIndexName(indexName, period)
                    res = self.existingIndex(actualIndexName)  # check actual index
                    if not res["OK"] or not res["Value"]:
                        result = self.createIndex(indexName, self.documentTypes[monitoringType]["mapping"], period)
                        if not result["OK"]:
                            self.log.error(result["Message"])
                            raise RuntimeError(result["Message"])
                        self.log.always("Index created", actualIndexName)
                else:
                    # in case the index pattern does not exist
                    result = self.createIndex(indexName, self.documentTypes[monitoringType]["mapping"], period)
                    if not result["OK"]:
                        self.log.error(result["Message"])
                        raise RuntimeError(result["Message"])
                    self.log.always("Index created", indexName)

    def getIndexName(self, typeName):
        """
        :param str typeName: monitoring type
        """
        indexName = None

        if typeName in self.documentTypes:
            indexName = self.documentTypes.get(typeName).get("indexName", None)

        if indexName:
            return S_OK(indexName)

        return S_ERROR("Monitoring type %s is not defined" % typeName)

    def getKeyValues(self, monitoringType):
        """
        Get all values for a given key field in a type
        """
        keyValuesDict = {}

        retVal = self.getIndexName(monitoringType)
        if not retVal["OK"]:
            return retVal
        indexName = "%s*" % (retVal["Value"])
        retVal = self.getDocTypes(indexName)
        if not retVal["OK"]:
            return retVal
        docs = retVal["Value"]
        self.log.debug("Doc types", docs)
        monfields = self.documentTypes[monitoringType]["monitoringFields"]

        try:
            properties = docs[monitoringType]["properties"]  # "old" way, with ES types == Monitoring types
        except KeyError:
            try:
                properties = docs["_doc"]["properties"]  # "ES6" way, with ES types == '_doc'
            except KeyError:
                properties = docs["properties"]  # "ES7" way, no types

        for i in properties:
            if i not in monfields and not i.startswith("time") and i != "metric":
                retVal = self.getUniqueValue(indexName, i)
                if not retVal["OK"]:
                    return retVal
                keyValuesDict[i] = retVal["Value"]
        return S_OK(keyValuesDict)

    def retrieveBucketedData(
        self, typeName, startTime, endTime, interval, selectField, condDict, grouping, metainfo=None
    ):
        """
        Get bucketed data from the DB. This is the standard method used.

        :param str typeName: name of the monitoring type
        :param int startTime: start time, expressed as epoch
        :param int endtime: end time, expressed as epoch
        :param str interval: interval expressed in ES notation
        :param str selectField: the field on which we are aggregating (the plot type)
        :param dict condDict: optional additional conditions for the query
        :param str grouping: what we are grouping on
        :param dict metainfo: meta information
        :returns: S_OK/S_ERROR with dictionary of key/value pairs
        """

        if not selectField:
            return S_ERROR("Missing the selection field")

        if not grouping:
            return S_ERROR("Missing the grouping field")

        aggName = "_".join([selectField, grouping])

        if metainfo is None:
            metainfo = {}

        isAvgAgg = False
        # the data is used to fill the pie charts.
        # This aggregation is used to average the buckets.
        if metainfo.get("metric", "sum") == "avg":
            isAvgAgg = True

        # Getting the index
        retVal = self.getIndexName(typeName)
        if not retVal["OK"]:
            return retVal
        indexName = "%s*" % (retVal["Value"])
        s = self._Search(indexName)

        # ## building the ES query incrementally

        # 2 parts:
        #
        # 1) query (on which documents are we working on?)
        # 2) aggregations (given the documents, on which field are we aggregating?)

        # 1) create the query to filter the results
        # Time range is always present
        q = [self._Q("range", timestamp={"lte": endTime * 1000, "gte": startTime * 1000})]
        # Optional additional conditions
        for cond in condDict:
            query = None
            for condValue in condDict[cond]:
                if not (cond and condValue):
                    continue
                kwargs = {cond: condValue}
                if query:
                    query = query | self._Q("term", **kwargs)
                else:
                    query = self._Q("term", **kwargs)
            if query:
                q += [query]

        s = s.filter("bool", must=q)

        # 2) prepare the aggregations

        # This the time-based metric aggregation of the selected field
        timeAggregation = self._A("terms", field="timestamp")
        timeAggregation.metric("total", "sum", field=selectField)  # name  # type

        # and now we group with bucket aggregation
        groupingAggregation = self._A("terms", field=grouping, size=self.RESULT_SIZE)
        groupingAggregation.bucket(
            "end_data", "date_histogram", field="timestamp", interval=interval  # name  # type
        ).metric("timeAggregation", timeAggregation).pipeline(
            "timeAggregation_avg_bucket", "avg_bucket", buckets_path="timeAggregation>total", gap_policy="insert_zeros"
        )
        if isAvgAgg:
            groupingAggregation.pipeline(
                "avg_total_jobs",
                "avg_bucket",
                buckets_path="end_data>timeAggregation_avg_bucket",
                gap_policy="insert_zeros",
            )

        s.aggs.bucket(aggName, groupingAggregation)

        s = s.source(False)  # don't return any fields (hits), just the metadata
        s = s[0:0]  # pagination 0, as we are only interested in the aggregations
        self.log.debug("Query:", s.to_dict())
        retVal = s.execute()

        self.log.debug("Query len and result", (str(len(retVal)) + " : " + str(retVal)))

        result = {}
        for bucket in retVal.aggregations[aggName].buckets:
            if isAvgAgg:
                if bucket.key not in result:
                    result[bucket.key] = bucket.avg_total_jobs.value
                else:
                    result[bucket.key].append(bucket.avg_total_jobs.value)
            else:
                if bucket.key not in result:
                    result[bucket.key] = {}
                for k in bucket.end_data.buckets:
                    if (k.key / 1000) not in result[bucket.key]:
                        result[bucket.key][k.key / 1000] = k.timeAggregation_avg_bucket.value
                    else:
                        result[bucket.key][k.key / 1000].append(k.timeAggregation_avg_bucket.value)

        return S_OK(result)

    def retrieveAggregatedData(
        self, typeName, startTime, endTime, interval, selectField, condDict, grouping, metainfo={}
    ):
        """
        Get data from the DB using simple aggregations.
        Note: this method is equivalent to retrieveBucketedData.
        The difference is in the dynamic bucketing.
        We do not perform dynamic bucketing on the raw data.

        :param str typeName: name of the monitoring type
        :param int startTime: start time, expressed as epoch
        :param int endtime: end time, expressed as epoch
        :param str interval: interval expressed in ES notation
        :param str selectField: the field on which we are aggregating (the plot type)
        :param dict condDict: optional additional conditions for the query
        :param str grouping: what we are grouping on
        :param dict metainfo: meta information
        :returns: S_OK/S_ERROR with dictionary of key/value pairs

        """
        # {'query':
        #     {'bool':
        #         {'filter': [
        #             {'bool':
        #                 {'must': [
        #                     {'range':
        #                         {'timestamp':
        #                             {'gte': 1474271462000, 'lte': 1474357862000}}}]}}]}},
        #     'aggs':
        #         {'end_data':
        #             {'date_histogram':
        #                 {'field': 'timestamp', 'interval': '30m'},
        #              'aggs': {'tt': {'terms': {'field': 'component', 'size': 10000},
        #                       'aggs': {'m1': {'avg': {'field': 'threads'}}}}}}}}

        # query = [Q('range', timestamp={'lte':1474357862000,'gte': 1474271462000} )]

        # a = A('terms', field='component', size=10000)
        # a.metric('m1', 'avg', field='threads')

        # s = Search(using=cl, index='lhcb-certification_componentmonitoring-index-*')

        # s = s.filter( 'bool', must = query )
        # s = s.aggs.bucket('end_data', 'date_histogram', field='timestamp', interval='30m').metric( 'tt', a )

        if not selectField:
            return S_ERROR("Missing the selection field")

        if not grouping:
            return S_ERROR("Missing the grouping field")

        if metainfo is None:
            metainfo = {}

        # Getting the index
        retVal = self.getIndexName(typeName)
        if not retVal["OK"]:
            return retVal
        indexName = "%s*" % (retVal["Value"])
        s = self._Search(indexName)

        # ## building the ES query incrementally

        # 2 parts:
        #
        # 1) query (on which documents are we working on?)
        # 2) aggregations (given the documents, on which field are we aggregating?)

        # 1) create the query to filter the results
        # Time range is always present
        q = [self._Q("range", timestamp={"lte": endTime * 1000, "gte": startTime * 1000})]
        # Optional additional conditions
        for cond in condDict:
            query = None
            for condValue in condDict[cond]:
                if not (cond and condValue):
                    continue
                kwargs = {cond: condValue}
                if query:
                    query = query | self._Q("term", **kwargs)
                else:
                    query = self._Q("term", **kwargs)
            if query:
                q += [query]
        s = s.filter("bool", must=q)

        # 2) prepare the aggregations

        a1 = self._A("terms", field=grouping, size=self.RESULT_SIZE)
        # default is average
        aggregator = metainfo.get("metric", "avg")
        a1.metric("m1", aggregator, field=selectField)
        s.aggs.bucket("end_data", "date_histogram", field="timestamp", interval=interval).metric("tt", a1)

        s = s.extra(size=self.RESULT_SIZE)  # max size
        s = s.source(False)  # don't return any fields (hits), just the metadata
        s = s[0:0]  # pagination 0, as we are only interested in the aggregations
        self.log.debug("Query:", s.to_dict())
        retVal = s.execute()

        result = {}
        for bucket in retVal.aggregations["end_data"].buckets:
            # each bucket key is a time (unix epoch and usual datetime
            bucketTime = int(bucket.key / 1000)
            for value in bucket["tt"].buckets:
                # each bucket contains an agregation called tt which sum/avg of the metric.
                if value.key not in result:
                    result[value.key] = {
                        bucketTime: value.m1.value if value.m1.value else 0
                    }  # TODO: this is kind of hack.
                    # we can use a default value for pipeline aggregation. EL promised that we
                    # can use default value for simple aggregation. Later to be checked.
                else:
                    result[value.key].update({bucketTime: value.m1.value if value.m1.value else 0})

        return S_OK(result)

    def put(self, records, monitoringType):
        """
        It is used to insert the data to ES. Calls bulk_index()

        :param list records: it is a list of documents (dictionary)
        :param str monitoringType: is the type of the monitoring
        """
        mapping = self.getMapping(monitoringType)
        self.log.verbose("Mapping used to create an index:", mapping)
        period = self.documentTypes[monitoringType].get("period")
        res = self.getIndexName(monitoringType)
        if not res["OK"]:
            return res
        indexName = res["Value"]

        return self.bulk_index(indexPrefix=indexName, data=records, mapping=mapping, period=period)

    def getMapping(self, monitoringType):
        """
        It returns the mapping of a certain monitoring type

        :param str monitoringType: the monitoring type for example WMSHistory
        :return: an empty dictionary if there is no mapping defenied.
        """
        mapping = {}
        if monitoringType in self.documentTypes:
            mapping = self.documentTypes[monitoringType].get("mapping", {})
        return mapping

    def __getRawData(self, typeName, condDict, size=-1):
        """
        It returns the last day data for a given monitoring type.
        :returns: for example

          .. code-block:: python

           {'sort': [{'timestamp': {'order': 'desc'}}],
            'query': {'bool': {'must': [{'match': {'host': 'dzmathe.cern.ch'}},
                                        {'match': {'component': 'Bookkeeping_BookkeepingManager'}}
                                       ]}}}

        :param str typeName: name of the monitoring type
        :param dict condDict: conditions for the query

                       * key -> name of the field
                       * value -> list of possible values

        :param int size: number of rows which whill be returned. By default is all
        """
        if size < 0:
            size = self.RESULT_SIZE
        retVal = self.getIndexName(typeName)
        if not retVal["OK"]:
            return retVal
        indexName = "%s-%s" % (retVal["Value"], time.strftime("%Y-%m-%d", time.gmtime()))

        # going to create:
        # s = Search(using=cl, index = 'lhcb-certification_componentmonitoring-index-2016-09-16')
        # s = s.filter( 'bool', must = [Q('match', host='dzmathe.cern.ch'),
        #  Q('match', component='Bookkeeping_BookkeepingManager')])
        # s = s.query(q)
        # s = s.sort('-timestamp')

        mustClose = []
        for cond in condDict:
            if cond not in ("startTime", "endTime"):
                kwargs = {cond: condDict[cond]}
                query = self._Q("match", **kwargs)
                mustClose.append(query)

        if condDict.get("startTime") and condDict.get("endTime"):
            query = self._Q(
                "range", timestamp={"lte": condDict.get("endTime") * 1000, "gte": condDict.get("startTime") * 1000}
            )

            mustClose.append(query)

        s = self._Search(indexName)
        s = s.filter("bool", must=mustClose)
        s = s.sort("-timestamp")

        if size > 0:
            s = s.extra(size=size)

        self.log.debug("Query:", s.to_dict())
        retVal = s.execute()
        if not retVal:
            self.log.error("Error getting raw data", str(retVal))
            return S_ERROR(str(retVal))
        hits = retVal["hits"]
        if hits and "hits" in hits and hits["hits"]:
            records = []
            paramNames = dir(hits["hits"][0]["_source"])
            try:
                paramNames.remove(u"metric")
            except KeyError as e:
                self.log.warn("metric is not in the Result", e)
            for resObj in hits["hits"]:
                records.append(dict([(paramName, getattr(resObj["_source"], paramName)) for paramName in paramNames]))
            return S_OK(records)

    def getLastDayData(self, typeName, condDict):
        """
        It returns the last day data for a given monitoring type.

        :returns: for example

          .. code-block:: python

           {'sort': [{'timestamp': {'order': 'desc'}}],
            'query': {'bool': {'must': [{'match': {'host': 'dzmathe.cern.ch'}},
                                        {'match': {'component': 'Bookkeeping_BookkeepingManager'}}]}}}

        :param str typeName: name of the monitoring type
        :param dict condDict: conditions for the query

                      * key -> name of the field
                      * value -> list of possible values
        """
        return self.__getRawData(typeName, condDict)

    def getLimitedData(self, typeName, condDict, size=10):
        """
        Returns a list of records for a given selection.

        :param str typeName: name of the monitoring type
        :param dict condDict: -> conditions for the query

                      * key -> name of the field
                      * value -> list of possible values

        :param int size: Indicates how many entries should be retrieved from the log
        :return: Up to size entries for the given component from the database
        """
        return self.__getRawData(typeName, condDict, size)

    def getDataForAGivenPeriod(self, typeName, condDict, initialDate="", endDate=""):
        """
        Retrieves the history of logging entries for the given component during a given given time period

        :param str typeName: name of the monitoring type
        :param dict condDict: conditions for the query

                      * key -> name of the field
                      * value -> list of possible values

        :param str initialDate: Indicates the start of the time period in the format 'DD/MM/YYYY hh:mm'
        :param str endDate: Indicate the end of the time period in the format 'DD/MM/YYYY hh:mm'
        :return: Entries from the database for the given component recorded between the initial and the end dates

        """
        if not initialDate and not endDate:
            return self.__getRawData(typeName, condDict, 10)

        if initialDate:
            condDict["startTime"] = calendar.timegm(time.strptime(initialDate, "%d/%m/%Y %H:%M"))
        else:
            condDict["startTime"] = 0000000000
        if endDate:
            condDict["endTime"] = calendar.timegm(time.strptime(endDate, "%d/%m/%Y %H:%M"))
        else:
            condDict["endTime"] = calendar.timegm(time.gmtime())

        return self.__getRawData(typeName, condDict)



#!/usr/bin/env python
"""
Module created to run failed jobs locally on a CVMFS-configured machine.
It creates the necessary environment, downloads the necessary files, modifies the necessary
files and runs the job

Usage:
  dirac-production-runjoblocal [Data imput mode] [job ID]

Arguments:
  Download (Job ID): Defines data aquisition as DownloadInputData
  Protocol (Job ID): Defines data acquisition as InputDataByProtocol
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import os
import shutil
import ssl

from six.moves.urllib.request import urlopen

from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


def __runSystemDefaults(jobID, vo):
    """
    Creates the environment for running the job and returns
    the path for the other functions.

    """
    from DIRAC.Core.Utilities.File import mkDir

    tempdir = str(vo) + "job" + str(jobID) + "temp"
    mkDir(tempdir)
    basepath = os.getcwd()
    return basepath + os.path.sep + tempdir + os.path.sep


def __downloadJobDescriptionXML(jobID, basepath):
    """
    Downloads the jobDescription.xml file into the temporary directory
    created.

    """
    from DIRAC.Interfaces.API.Dirac import Dirac

    jdXML = Dirac()
    jdXML.getInputSandbox(jobID, basepath)


def __modifyJobDescription(jobID, basepath, downloadinputdata):
    """
    Modifies the jobDescription.xml to, instead of DownloadInputData, it
    uses InputDataByProtocol

    """
    from DIRAC import S_OK

    if not downloadinputdata:
        from xml.etree import ElementTree as et

        archive = et.parse(basepath + "InputSandbox" + str(jobID) + os.path.sep + "jobDescription.xml")
        for element in archive.iter():
            if element.text == "DIRAC.WorkloadManagementSystem.Client.DownloadInputData":
                element.text = "DIRAC.WorkloadManagementSystem.Client.InputDataByProtocol"
                archive.write(basepath + "InputSandbox" + str(jobID) + os.path.sep + "jobDescription.xml")


def __downloadPilotScripts(basepath, diracpath):
    """
    Downloads the scripts necessary to configure the pilot

    """

    context = ssl._create_unverified_context()
    for fileName in [
        "dirac-pilot.py",
        "pilotCommands.py",
        "pilotTools",
        "MessageSender",
        "PilotLogger.py",
        "PilotLoggerTools.py",
    ]:
        remoteFile = urlopen(
            os.path.join("https://raw.githubusercontent.com/DIRACGrid/Pilot/master/Pilot/", fileName),
            timeout=10,
            context=context,
        )
        with open(fileName, "wb") as localFile:
            localFile.write(remoteFile.read())
    diracInstall = urlopen(
        "https://raw.githubusercontent.com/DIRACGrid/management/master/dirac-install.py ", timeout=10, context=context
    )
    with open("dirac-install.py", "wb") as localFile:
        localFile.write(diracInstall.read())


def __configurePilot(basepath, vo):
    """
    Configures the pilot.
    This method was created specifically for LHCb pilots, more info
    about othe VOs is needed to make it more general.
    """

    from DIRAC.ConfigurationSystem.Client.Helpers.CSGlobals import getVO, getSetup
    from DIRAC.ConfigurationSystem.Client.ConfigurationData import gConfigurationData

    vo = getVO()
    currentSetup = getSetup()
    masterCS = gConfigurationData.getMasterServer()

    os.system(
        "python "
        + basepath
        + "dirac-pilot.py -S %s -l %s -C %s -N ce.debug.ch -Q default -n DIRAC.JobDebugger.ch -dd"
        % (currentSetup, vo, masterCS)
    )

    diracdir = os.path.expanduser("~") + os.path.sep
    try:
        os.rename(diracdir + ".dirac.cfg", diracdir + ".dirac.cfg.old")
    except OSError:
        pass
    shutil.copyfile(diracdir + "pilot.cfg", diracdir + ".dirac.cfg")


def __runJobLocally(jobID, basepath, vo):
    """
    Runs the job!

    """
    ipr = __import__(str(vo) + "DIRAC.Interfaces.API." + str(vo) + "Job", globals(), locals(), [str(vo) + "Job"], -1)
    voJob = getattr(ipr, str(vo) + "Job")
    localJob = voJob(basepath + "InputSandbox" + str(jobID) + os.path.sep + "jobDescription.xml")
    localJob.setInputSandbox(os.getcwd() + os.path.sep + "pilot.cfg")
    localJob.setConfigArgs(os.getcwd() + os.path.sep + "pilot.cfg")
    os.chdir(basepath)
    localJob.runLocal()


@Script()
def main():
    Script.registerSwitch("D:", "Download=", "Defines data acquisition as DownloadInputData")
    Script.registerSwitch("P:", "Protocol=", "Defines data acquisition as InputDataByProtocol")
    Script.parseCommandLine(ignoreErrors=False)

    _downloadinputdata = False
    _jobID = None

    for switch in Script.getUnprocessedSwitches():
        if switch[0] in ("D", "Download"):
            _downloadinputdata = True
            _jobID = switch[1]
        if switch[0] in ("I", "Protocol"):
            _downloadinputdata = False
            _jobID = switch[1]

    from DIRAC.ConfigurationSystem.Client.Helpers.CSGlobals import Extensions

    ext = Extensions()
    _vo = ext.getCSExtensions()[0]
    _diracPath = Extensions().getExtensionPath("DIRAC")
    _dir = os.path.expanduser("~") + os.path.sep
    try:
        _path = __runSystemDefaults(_jobID, _vo)

        __downloadJobDescriptionXML(_jobID, _path)

        __modifyJobDescription(_jobID, _path, _downloadinputdata)

        __downloadPilotScripts(_path, _diracPath)

        __configurePilot(_path, _vo)

        __runJobLocally(_jobID, _path, _vo)

    finally:
        os.chdir(_dir)
        os.rename(_dir + ".dirac.cfg.old", _dir + ".dirac.cfg")


if __name__ == "__main__":
    main()



""" Test class for LSFResourceUsage utility
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import pytest

from DIRAC import S_OK, S_ERROR

from DIRAC.Resources.Computing.BatchSystems.TimeLeft.LSFResourceUsage import LSFResourceUsage

LSF_KEK_BQUEUES = """   CPULIMIT
  720.0 min

  RUNLIMIT
  1440.0 min
"""

LSF_LSHOSTS = """ HOST_NAME                       type       model  cpuf ncpus maxmem maxswp server RESOURCES
b66                   SLC6_64 i6_16   2.5    16 29M 19M    Yes (intel share aishare cvmfs wan exe lcg wigner slot15)
  """

LSF_CERN_BQUEUES = """   CPULIMIT
  10080.0 min of KSI2K

  RUNLIMIT
  30240.0 min of KSI2K
"""

# returns with S_ERROR
LSF_CERN_LSHOSTS_1 = """KSI2K: unknown host name.
"""

# shortened
LSF_CERN_LSINFO = """MODEL_NAME      CPU_FACTOR      ARCHITECTURE
i6_12_62d7h20_266      3.06
ai_intel_8            2.44
"""


@pytest.mark.parametrize(
    "runCommandResult, lsbHosts, cpuLimitExpected, wallClockLimitExpected",
    [((S_OK(LSF_KEK_BQUEUES), S_OK(LSF_LSHOSTS)), "b66", (720 * 60 / 2.5), (1440 * 60 / 2.5))],
)
def test_init(mocker, runCommandResult, lsbHosts, cpuLimitExpected, wallClockLimitExpected):
    mocker.patch(
        "DIRAC.Resources.Computing.BatchSystems.TimeLeft.LSFResourceUsage.runCommand", side_effect=runCommandResult
    )
    mocker.patch.dict(os.environ, {"LSB_HOSTS": lsbHosts})

    lsfResourceUsage = LSFResourceUsage()
    assert lsfResourceUsage.cpuLimit == cpuLimitExpected
    assert lsfResourceUsage.wallClockLimit == wallClockLimitExpected


@pytest.mark.parametrize(
    "runCommandResult, sourceEnvResult, lsbHosts, lsfEnvdir, cpuLimitExpected, \
                         wallClockLimitExpected, normrefExpected, hostnormExpected, cpuRef",
    [
        (
            (S_OK(LSF_CERN_BQUEUES), S_ERROR(LSF_CERN_LSHOSTS_1), S_OK(LSF_CERN_LSINFO), S_OK(LSF_LSHOSTS)),
            S_ERROR("no lsf.sh"),
            "b66",
            "/dev/null",
            241920,
            725760,
            1.0,
            2.5,
            "KSI2K",
        )
    ],
)
def test_init_cern(
    mocker,
    runCommandResult,
    sourceEnvResult,
    lsbHosts,
    lsfEnvdir,
    cpuLimitExpected,
    wallClockLimitExpected,
    normrefExpected,
    hostnormExpected,
    cpuRef,
):
    mocker.patch(
        "DIRAC.Resources.Computing.BatchSystems.TimeLeft.LSFResourceUsage.runCommand", side_effect=runCommandResult
    )
    mocker.patch("os.path.isfile", return_value=True)
    mocker.patch.dict(os.environ, {"LSB_HOSTS": lsbHosts, "LSF_ENVDIR": lsfEnvdir})
    mocker.patch(
        "DIRAC.Resources.Computing.BatchSystems.TimeLeft.LSFResourceUsage.sourceEnv", return_value=sourceEnvResult
    )

    lsfResourceUsage = LSFResourceUsage()

    assert lsfResourceUsage.cpuLimit == cpuLimitExpected
    assert lsfResourceUsage.wallClockLimit == wallClockLimitExpected
    assert lsfResourceUsage.cpuRef == cpuRef
    assert lsfResourceUsage.normRef == normrefExpected
    assert lsfResourceUsage.hostNorm == hostnormExpected



""" :mod: RucioRSSAgent

    Agent that synchronizes Rucio and Dirac


The following options can be set for the RucioRSSAgent.

.. literalinclude:: ../ConfigTemplate.cfg
  :start-after: ##BEGIN RucioRSSAgent
  :end-before: ##END
  :dedent: 2
  :caption: RucioRSSAgent options
"""

# # imports
from DIRAC import S_OK, S_ERROR
from DIRAC.Core.Base.AgentModule import AgentModule
from DIRAC.ResourceStatusSystem.Client.ResourceStatus import ResourceStatus

from rucio.client import Client

__RCSID__ = "Id$"


class RucioRSSAgent(AgentModule):
    """
    .. class:: RucioRSSAgent

    Agent that synchronizes Rucio and Dirac
    """

    def execute(self):
        """execution in one agent's cycle

        :param self: self reference
        """
        rSS = ResourceStatus()
        client = Client()
        try:
            for rse in client.list_rses():
                thisSe = rse["rse"]
                self.log.info("Running on %s", thisSe)
                resStatus = rSS.getElementStatus(thisSe, "StorageElement")
                dictSe = client.get_rse(thisSe)
                if resStatus["OK"]:
                    seAccessValue = resStatus["Value"][thisSe]
                    availabilityRead = True if seAccessValue["ReadAccess"] in ["Active", "Degraded"] else False
                    availabilityWrite = True if seAccessValue["WriteAccess"] in ["Active", "Degraded"] else False
                    availabilityDelete = True if seAccessValue["RemoveAccess"] in ["Active", "Degraded"] else False
                    isUpdated = False
                    if dictSe["availability_read"] != availabilityRead:
                        self.log.info("Set availability_read for %s to %s", thisSe, availabilityRead)
                        client.update_rse(thisSe, {"availability_read": availabilityRead})
                        isUpdated = True
                    if dictSe["availability_write"] != availabilityWrite:
                        self.log.info("Set availability_write for %s to %s", thisSe, availabilityWrite)
                        client.update_rse(thisSe, {"availability_write": availabilityWrite})
                        isUpdated = True
                    if dictSe["availability_delete"] != availabilityDelete:
                        self.log.info("Set availability_delete for %s to %s", thisSe, availabilityDelete)
                        client.update_rse(thisSe, {"availability_delete": availabilityDelete})
                        isUpdated = True
        except Exception as err:
            return S_ERROR(str(err))
        return S_OK()



""" SystemAdministrator service is a tool to control and monitor the DIRAC services and agents
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import socket
import os
import re
import time
import getpass
import importlib
import shutil
import platform
import psutil
import tempfile

import six
from packaging.version import Version, InvalidVersion

try:
    import subprocess32 as subprocess
except ImportError:
    import subprocess

# TODO: This should be modernised to use subprocess(32)
try:
    import commands
except ImportError:
    # Python 3's subprocess module contains a compatibility layer
    import subprocess as commands

from datetime import datetime, timedelta
from distutils.spawn import find_executable

from diraccfg import CFG
import requests

from DIRAC import S_OK, S_ERROR, gConfig, rootPath, gLogger, convertToPy3VersionNumber
from DIRAC.Core.DISET.RequestHandler import RequestHandler
from DIRAC.Core.Utilities import Os
from DIRAC.Core.Utilities.Extensions import extensionsByPriority, getExtensionMetadata
from DIRAC.Core.Utilities.File import mkLink
from DIRAC.Core.Utilities.Time import dateTime, fromString, hour, day
from DIRAC.Core.Utilities.Subprocess import shellCall, systemCall
from DIRAC.Core.Utilities.ThreadScheduler import gThreadScheduler
from DIRAC.Core.Security.Locations import getHostCertificateAndKeyLocation
from DIRAC.Core.Security.X509Chain import X509Chain  # pylint: disable=import-error
from DIRAC.ConfigurationSystem.Client import PathFinder
from DIRAC.ConfigurationSystem.Client.Helpers.CSGlobals import getCSExtensions
from DIRAC.FrameworkSystem.Client.ComponentInstaller import gComponentInstaller
from DIRAC.FrameworkSystem.Client.ComponentMonitoringClient import ComponentMonitoringClient

gProfilers = {}

# pylint: disable=no-self-use


def loadDIRACCFG():
    installPath = gConfig.getValue("/LocalInstallation/TargetPath", gConfig.getValue("/LocalInstallation/RootPath", ""))
    if not installPath:
        installPath = rootPath
    cfgPath = os.path.join(installPath, "etc", "dirac.cfg")
    try:
        diracCFG = CFG().loadFromFile(cfgPath)
    except Exception as excp:
        return S_ERROR("Could not load dirac.cfg: %s" % repr(excp))

    return S_OK((cfgPath, diracCFG))


class SystemAdministratorHandler(RequestHandler):
    @classmethod
    def initializeHandler(cls, serviceInfo):
        """
        Handler class initialization
        """

        # Check the flag for monitoring of the state of the host
        hostMonitoring = cls.srv_getCSOption("HostMonitoring", True)

        if hostMonitoring:
            gThreadScheduler.addPeriodicTask(60, cls.__storeHostInfo)
            # the SystemAdministrator service does not has to use the client to report data about the host.

        keepSoftwareVersions = cls.srv_getCSOption("KeepSoftwareVersions", 0)
        if keepSoftwareVersions > 0:
            gLogger.info(
                "The last %s software version will be kept and the rest will be deleted!" % keepSoftwareVersions
            )
            gThreadScheduler.addPeriodicTask(
                600, cls.__deleteOldSoftware, (keepSoftwareVersions,), executions=2
            )  # it is enough to try 2 times

        return S_OK("Initialization went well")

    types_getInfo = []

    def export_getInfo(self):
        """Get versions of the installed DIRAC software and extensions, setup of the
        local installation
        """
        return gComponentInstaller.getInfo()

    types_getSoftwareComponents = []

    def export_getSoftwareComponents(self):
        """Get the list of all the components ( services and agents ) for which the software
        is installed on the system
        """
        return gComponentInstaller.getSoftwareComponents(extensionsByPriority())

    types_getInstalledComponents = []

    def export_getInstalledComponents(self):
        """Get the list of all the components ( services and agents )
        installed on the system in the runit directory
        """
        return gComponentInstaller.getInstalledComponents()

    types_getSetupComponents = []

    def export_getSetupComponents(self):
        """Get the list of all the components ( services and agents )
        set up for running with runsvdir in /opt/dirac/startup directory
        """
        return gComponentInstaller.getSetupComponents()

    types_getOverallStatus = []

    def export_getOverallStatus(self):
        """Get the complete status information of all components."""
        result = gComponentInstaller.getOverallStatus(extensionsByPriority())
        if not result["OK"]:
            return result
        statusDict = result["Value"]
        for compType in statusDict:
            for system in statusDict[compType]:
                for component in statusDict[compType][system]:
                    result = gComponentInstaller.getComponentModule(system, component, compType)
                    if not result["OK"]:
                        statusDict[compType][system][component]["Module"] = "Unknown"
                    else:
                        statusDict[compType][system][component]["Module"] = result["Value"]
        return S_OK(statusDict)

    types_getStartupComponentStatus = [list]

    def export_getStartupComponentStatus(self, componentTupleList):
        """Get the list of all the components ( services and agents )
        set up for running with runsvdir in startup directory
        """
        return gComponentInstaller.getStartupComponentStatus(componentTupleList)

    types_installComponent = [six.string_types, six.string_types, six.string_types]

    def export_installComponent(self, componentType, system, component, componentModule=""):
        """Install runit directory for the specified component"""
        return gComponentInstaller.installComponent(
            componentType, system, component, extensionsByPriority(), componentModule
        )

    types_setupComponent = [six.string_types, six.string_types, six.string_types]

    def export_setupComponent(self, componentType, system, component, componentModule=""):
        """Setup the specified component for running with the runsvdir daemon
        It implies installComponent
        """
        result = gComponentInstaller.setupComponent(
            componentType, system, component, extensionsByPriority(), componentModule
        )
        gConfig.forceRefresh()
        return result

    types_addDefaultOptionsToComponentCfg = [six.string_types, six.string_types]

    def export_addDefaultOptionsToComponentCfg(self, componentType, system, component):
        """Add default component options local component cfg"""
        return gComponentInstaller.addDefaultOptionsToComponentCfg(
            componentType, system, component, extensionsByPriority()
        )

    types_unsetupComponent = [six.string_types, six.string_types]

    def export_unsetupComponent(self, system, component):
        """Removed the specified component from running with the runsvdir daemon"""
        return gComponentInstaller.unsetupComponent(system, component)

    types_uninstallComponent = [six.string_types, six.string_types, bool]

    def export_uninstallComponent(self, system, component, removeLogs):
        """Remove runit directory for the specified component
        It implies unsetupComponent
        """
        return gComponentInstaller.uninstallComponent(system, component, removeLogs)

    types_startComponent = [six.string_types, six.string_types]

    def export_startComponent(self, system, component):
        """Start the specified component, running with the runsv daemon"""
        return gComponentInstaller.runsvctrlComponent(system, component, "u")

    types_restartComponent = [six.string_types, six.string_types]

    def export_restartComponent(self, system, component):
        """Restart the specified component, running with the runsv daemon"""
        return gComponentInstaller.runsvctrlComponent(system, component, "t")

    types_stopComponent = [six.string_types, six.string_types]

    def export_stopComponent(self, system, component):
        """Stop the specified component, running with the runsv daemon"""
        return gComponentInstaller.runsvctrlComponent(system, component, "d")

    types_getLogTail = [six.string_types, six.string_types]

    def export_getLogTail(self, system, component, length=100):
        """Get the tail of the component log file"""
        return gComponentInstaller.getLogTail(system, component, length)

    ######################################################################################
    #  Database related methods
    #
    types_getMySQLStatus = []

    def export_getMySQLStatus(self):
        """Get the status of the MySQL database installation"""
        return gComponentInstaller.getMySQLStatus()

    types_getDatabases = []

    def export_getDatabases(self, mysqlPassword=None):
        """Get the list of installed databases"""
        if mysqlPassword:
            gComponentInstaller.setMySQLPasswords(mysqlPassword)
        return gComponentInstaller.getDatabases()

    types_getAvailableDatabases = []

    def export_getAvailableDatabases(self):
        """Get the list of databases which software is installed in the system"""
        return gComponentInstaller.getAvailableDatabases()

    types_installDatabase = [six.string_types]

    def export_installDatabase(self, dbName, mysqlPassword=None):
        """Install a DIRAC database named dbName"""
        if mysqlPassword:
            gComponentInstaller.setMySQLPasswords(mysqlPassword)
        return gComponentInstaller.installDatabase(dbName)

    types_uninstallDatabase = [six.string_types]

    def export_uninstallDatabase(self, dbName, mysqlPassword=None):
        """Uninstall a DIRAC database named dbName"""
        if mysqlPassword:
            gComponentInstaller.setMySQLPasswords(mysqlPassword)
        return gComponentInstaller.uninstallDatabase(gConfig, dbName)

    types_addDatabaseOptionsToCS = [six.string_types, six.string_types]

    def export_addDatabaseOptionsToCS(self, system, database, overwrite=False):
        """Add the section with the database options to the CS"""
        return gComponentInstaller.addDatabaseOptionsToCS(gConfig, system, database, overwrite=overwrite)

    types_addDefaultOptionsToCS = [six.string_types, six.string_types, six.string_types]

    def export_addDefaultOptionsToCS(self, componentType, system, component, overwrite=False):
        """Add default component options to the global CS or to the local options"""
        return gComponentInstaller.addDefaultOptionsToCS(
            gConfig, componentType, system, component, extensionsByPriority(), overwrite=overwrite
        )

    #######################################################################################
    # General purpose methods
    #
    types_updateSoftware = [six.string_types]

    def export_updateSoftware(self, version, rootPath="", diracOSVersion=""):
        if "." in version:
            return self._updateSoftwarePy3(version, rootPath, diracOSVersion)
        else:
            return self._updateSoftwarePy2(version, rootPath, diracOSVersion)

    def _updateSoftwarePy3(self, version, rootPath_, diracOSVersion):
        if rootPath_:
            return S_ERROR("rootPath argument is not supported for Python 3 installations")
        # Validate and normalise the requested version
        primaryExtension = None
        if "==" in version:
            primaryExtension, version = version.split("==")
        elif six.PY2:
            return S_ERROR(
                "The extension must be specified like DIRAC==vX.Y.Z if installing "
                "a Python 3 version from an existing Python 2 installation."
            )
        try:
            version = Version(version)
        except InvalidVersion:
            self.log.exception("Invalid version passed", version)
            return S_ERROR("Invalid version passed %r" % version)
        isPrerelease = version.is_prerelease
        version = "v%s" % version

        # Find what to install
        otherExtensions = []
        for extension in extensionsByPriority():
            if primaryExtension is None and getExtensionMetadata(extension).get("primary_extension", False):
                primaryExtension = extension
            else:
                otherExtensions.append(extension)
        self.log.info(
            "Installing Python 3 based", "%s %s with DIRACOS %s" % (primaryExtension, version, diracOSVersion or "2")
        )
        self.log.info("Will also install", repr(otherExtensions))

        # Install DIRACOS
        installer_url = "https://github.com/DIRACGrid/DIRACOS2/releases/"
        if diracOSVersion:
            installer_url += "download/%s/latest/download/DIRACOS-Linux-%s.sh" % (version, platform.machine())
        else:
            installer_url += "latest/download/DIRACOS-Linux-%s.sh" % platform.machine()
        self.log.info("Downloading DIRACOS2 installer from", installer_url)
        with tempfile.NamedTemporaryFile(suffix=".sh", mode="wb") as installer:
            with requests.get(installer_url, stream=True) as r:
                if not r.ok:
                    return S_ERROR("Failed to download %s" % installer_url)
                for chunk in r.iter_content(chunk_size=1024 ** 2):
                    installer.write(chunk)
            installer.flush()
            self.log.info("Downloaded DIRACOS installer to", installer.name)

            if six.PY2:
                newProPrefix = os.path.dirname(os.path.realpath(os.path.join(rootPath, "bashrc")))
            else:
                newProPrefix = rootPath
            newProPrefix = os.path.join(
                newProPrefix,
                "versions",
                "%s-%s" % (version, datetime.utcnow().strftime("%s")),
            )
            installPrefix = os.path.join(newProPrefix, "%s-%s" % (platform.system(), platform.machine()))
            self.log.info("Running DIRACOS installer for prefix", installPrefix)
            r = subprocess.run(  # pylint: disable=no-member
                ["bash", installer.name, "-p", installPrefix],
                stderr=subprocess.PIPE,
                universal_newlines=True,
                check=False,
                timeout=300,
            )
            if r.returncode != 0:
                stderr = [x for x in r.stderr.split("\n") if not x.startswith("Extracting : ")]
                self.log.error(
                    "Installing DIRACOS2 failed with returncode", "%s and stdout: %s" % (r.returncode, stderr)
                )
                return S_ERROR("Failed to install DIRACOS2 %s" % stderr)

        # Install DIRAC
        cmd = ["%s/bin/pip" % installPrefix, "install", "--no-color", "-v"]
        if isPrerelease:
            cmd += ["--pre"]
        cmd += ["%s[server]==%s" % (primaryExtension, version)]
        cmd += ["%s[server]" % e for e in otherExtensions]
        r = subprocess.run(  # pylint: disable=no-member
            cmd,
            stderr=subprocess.PIPE,
            universal_newlines=True,
            check=False,
            timeout=600,
        )
        if r.returncode != 0:
            self.log.error("Installing DIRACOS2 failed with returncode", "%s and stdout: %s" % (r.returncode, r.stderr))
            return S_ERROR("Failed to install DIRACOS2 with message %s" % r.stderr)

        # Update the pro link
        oldLink = os.path.join(gComponentInstaller.instancePath, "old")
        proLink = os.path.join(gComponentInstaller.instancePath, "pro")
        if os.path.exists(oldLink):
            os.remove(oldLink)
        os.rename(proLink, oldLink)
        mkLink(newProPrefix, proLink)

        return S_OK()

    def _updateSoftwarePy2(self, version, rootPath, diracOSVersion):
        """Update the local DIRAC software installation to version"""
        # Check that we have a sane local configuration
        result = gConfig.getOptionsDict("/LocalInstallation")
        if not result["OK"]:
            return S_ERROR("Invalid installation - missing /LocalInstallation section in the configuration")
        elif not result["Value"]:
            return S_ERROR("Invalid installation - empty /LocalInstallation section in the configuration")

        if rootPath and not os.path.exists(rootPath):
            return S_ERROR('Path "%s" does not exists' % rootPath)

        installer = None
        if not find_executable("dirac-install"):
            installer = tempfile.NamedTemporaryFile(suffix=".py", mode="wb")
            with requests.get(
                "https://raw.githubusercontent.com/DIRACGrid/management/master/dirac-install.py", stream=True
            ) as r:
                if not r.ok:
                    return S_ERROR("Failed to download dirac-install from management repo")
                for chunk in r.iter_content(chunk_size=1024 ** 2):
                    installer.write(chunk)
            installer.flush()
            self.log.info("Downloaded dirac-install.py py2 installer to", installer.name)
            cmdList = ["python", installer.name, "-r", version, "-t", "server"]
        else:
            cmdList = ["dirac-install", "-r", version, "-t", "server"]

        if rootPath:
            cmdList.extend(["-P", rootPath])

        # Check if there are extensions
        extensionList = getCSExtensions()
        # By default we do not install WebApp
        if "WebApp" in extensionList or []:
            extensionList.remove("WebApp")

        webPortal = gConfig.getValue("/LocalInstallation/WebApp", False)
        if webPortal and "WebAppDIRAC" not in extensionList:
            extensionList.append("WebAppDIRAC")

        cmdList += ["-e", ",".join(extensionList)]

        project = gConfig.getValue("/LocalInstallation/Project")
        if project:
            cmdList += ["-l", project]

        targetPath = gConfig.getValue(
            "/LocalInstallation/TargetPath", gConfig.getValue("/LocalInstallation/RootPath", "")
        )
        if targetPath and os.path.exists(targetPath + "/etc/dirac.cfg"):
            cmdList.append(targetPath + "/etc/dirac.cfg")
        else:
            return S_ERROR("Local configuration not found")

        result = systemCall(240, cmdList)
        if installer:
            installer.close()
        if not result["OK"]:
            return result
        status = result["Value"][0]
        if status == 0:
            return S_OK()
        # Get error messages
        error = [line.strip() for line in result["Value"][1].split("\n") if "error" in line.lower()]
        return S_ERROR("\n".join(error or "Failed to update software to %s" % version))

    types_revertSoftware = []

    def export_revertSoftware(self):
        """Revert the last installed version of software to the previous one"""
        oldLink = os.path.join(gComponentInstaller.instancePath, "old")
        oldPath = os.readlink(oldLink)
        proLink = os.path.join(gComponentInstaller.instancePath, "pro")
        os.remove(proLink)
        mkLink(oldPath, proLink)

        return S_OK(oldPath)

    types_setProject = [six.string_types]

    def export_setProject(self, projectName):
        result = loadDIRACCFG()
        if not result["OK"]:
            return result
        cfgPath, diracCFG = result["Value"]
        gLogger.notice("Setting project to %s" % projectName)
        diracCFG.setOption("/LocalInstallation/Project", projectName, "Project to install")
        try:
            with open(cfgPath, "w") as fd:
                fd.write(str(diracCFG))
        except IOError as excp:
            return S_ERROR("Could not write dirac.cfg: %s" % str(excp))
        return S_OK()

    types_getProject = []

    def export_getProject(self):
        result = loadDIRACCFG()
        if not result["OK"]:
            return result
        _cfgPath, diracCFG = result["Value"]
        return S_OK(diracCFG.getOption("/LocalInstallation/Project", "DIRAC"))

    types_addOptionToDiracCfg = [six.string_types, six.string_types]

    def export_addOptionToDiracCfg(self, option, value):
        """Set option in the local configuration file"""
        return gComponentInstaller.addOptionToDiracCfg(option, value)

    types_executeCommand = [six.string_types]

    def export_executeCommand(self, command):
        """Execute a command locally and return its output"""
        result = shellCall(60, command)
        return result

    types_checkComponentLog = [[six.string_types, list]]

    def export_checkComponentLog(self, component):
        """Check component log for errors"""
        componentList = []
        if "*" in component:
            if component == "*":
                result = gComponentInstaller.getSetupComponents()
                if result["OK"]:
                    for ctype in ["Services", "Agents", "Executors"]:
                        if ctype in result["Value"]:
                            for sname in result["Value"][ctype]:
                                for cname in result["Value"][ctype][sname]:
                                    componentList.append("/".join([sname, cname]))
        elif isinstance(component, six.string_types):
            componentList = [component]
        else:
            componentList = component

        resultDict = {}
        for comp in componentList:
            if "/" not in comp:
                continue
            system, cname = comp.split("/")

            startDir = gComponentInstaller.startDir
            currentLog = startDir + "/" + system + "_" + cname + "/log/current"
            try:
                with open(currentLog, "r") as logFile:
                    logLines = logFile.readlines()
            except IOError as err:
                gLogger.error("File does not exists:", currentLog)
                resultDict[comp] = {"ErrorsHour": -1, "ErrorsDay": -1, "LastError": currentLog + "::" + repr(err)}
                continue

            errors_1 = 0
            errors_24 = 0
            now = dateTime()
            lastError = ""
            for line in logLines:
                if "ERROR:" in line:
                    fields = line.split()
                    recent = False
                    if len(fields) < 2:  # if the line contains only one word
                        lastError = line.split("ERROR:")[-1].strip()
                        continue
                    timeStamp = fromString(fields[0] + " " + fields[1])
                    if not timeStamp:  # if the timestamp is missing in the log
                        lastError = line.split("ERROR:")[-1].strip()
                        continue
                    if (now - timeStamp) < hour:
                        errors_1 += 1
                        recent = True
                    if (now - timeStamp) < day:
                        errors_24 += 1
                        recent = True
                    if recent:
                        lastError = line.split("ERROR:")[-1].strip()

            resultDict[comp] = {"ErrorsHour": errors_1, "ErrorsDay": errors_24, "LastError": lastError}

        return S_OK(resultDict)

    @staticmethod
    def __readHostInfo():
        """Get host current loads, memory, etc"""

        result = dict()
        # Memory info
        re_parser = re.compile(r"^(?P<key>\S*):\s*(?P<value>\d*)\s*kB")
        for line in open("/proc/meminfo"):
            match = re_parser.match(line)
            if not match:
                continue
            key, value = match.groups(["key", "value"])
            result[key] = int(value)

        for mtype in ["Mem", "Swap"]:
            memory = int(result.get(mtype + "Total"))
            mfree = int(result.get(mtype + "Free"))
            if memory > 0:
                percentage = float(memory - mfree) / float(memory) * 100.0
            else:
                percentage = 0
            name = "Memory"
            if mtype == "Swap":
                name = "Swap"
            result[name] = "%.1f%%/%.1fMB" % (percentage, memory / 1024.0)

        # Loads
        l1, l5, l15 = (str(lx) for lx in os.getloadavg())
        result["Load1"] = l1
        result["Load5"] = l5
        result["Load15"] = l15
        result["Load"] = "/".join([l1, l5, l15])

        # CPU info
        with open("/proc/cpuinfo", "r") as fd:
            lines = fd.readlines()
            processors = 0
            physCores = {}
            for line in lines:
                if line.strip():
                    parameter, value = line.split(":")
                    parameter = parameter.strip()
                    value = value.strip()
                    if parameter.startswith("processor"):
                        processors += 1
                    if parameter.startswith("physical id"):
                        physCores[value] = parameter
                    if parameter.startswith("model name"):
                        result["CPUModel"] = value
                    if parameter.startswith("cpu MHz"):
                        result["CPUClock"] = value
            result["Cores"] = processors
            result["PhysicalCores"] = len(physCores)

        # Disk occupancy
        summary = ""
        _status, output = commands.getstatusoutput("df")
        lines = output.split("\n")
        for i in range(len(lines)):
            if lines[i].startswith("/dev"):
                fields = lines[i].split()
                if len(fields) == 1:
                    fields += lines[i + 1].split()
                partition = fields[5]
                occupancy = fields[4]
                summary += ",%s:%s" % (partition, occupancy)
        result["DiskOccupancy"] = summary[1:]
        result["RootDiskSpace"] = Os.getDiskSpace(rootPath)

        # Open files
        puser = getpass.getuser()
        _status, output = commands.getstatusoutput("lsof")
        pipes = 0
        files = 0
        sockets = 0
        lines = output.split("\n")
        for line in lines:
            fType = line.split()[4]
            user = line.split()[2]
            if user == puser:
                if fType in ["REG"]:
                    files += 1
                elif fType in ["unix", "IPv4"]:
                    sockets += 1
                elif fType in ["FIFO"]:
                    pipes += 1
        result["OpenSockets"] = sockets
        result["OpenFiles"] = files
        result["OpenPipes"] = pipes

        infoResult = gComponentInstaller.getInfo()
        if infoResult["OK"]:
            result.update(infoResult["Value"])
            # the infoResult value is {"Extensions":{'a1':'v1',a2:'v2'}; we convert to a string
            result.update(
                {
                    "Extensions": ";".join(
                        ["%s:%s" % (key, value) for (key, value) in infoResult["Value"].get("Extensions").items()]
                    )
                }
            )

        # Host certificate properties
        certFile, _keyFile = getHostCertificateAndKeyLocation()
        chain = X509Chain()
        chain.loadChainFromFile(certFile)
        resultCert = chain.getCredentials()
        if resultCert["OK"]:
            result["SecondsLeft"] = resultCert["Value"]["secondsLeft"]
            result["CertificateValidity"] = str(timedelta(seconds=resultCert["Value"]["secondsLeft"]))
            result["CertificateDN"] = resultCert["Value"]["subject"]
            result["CertificateIssuer"] = resultCert["Value"]["issuer"]

        # Host uptime
        result["Uptime"] = str(timedelta(seconds=(time.time() - psutil.boot_time())))

        return S_OK(result)

    types_getHostInfo = []

    def export_getHostInfo(self):
        """
        Retrieve host parameters
        """
        client = ComponentMonitoringClient()
        result = client.getLog(socket.getfqdn())

        if result["OK"]:
            return S_OK(result["Value"][0])
        return self.__readHostInfo()

    types_getUsedPorts = []

    def export_getUsedPorts(self):
        """
        Retrieve the ports in use by services on this host

        :return: Returns a dictionary containing, for each system, which port is being used by which service
        """
        result = gComponentInstaller.getSetupComponents()
        if not result["OK"]:
            return result

        services = result["Value"]["Services"]
        ports = {}
        for system in services:
            ports[system] = {}
            for service in services[system]:
                url = PathFinder.getServiceURL("%s/%s" % (system, service))
                port = re.search(r":(\d{4,5})/", url)
                if port:
                    ports[system][service] = port.group(1)
                else:
                    ports[system][service] = "None"

        return S_OK(ports)

    types_getComponentDocumentation = [six.string_types, six.string_types, six.string_types]

    def export_getComponentDocumentation(self, cType, system, module):
        if cType == "service":
            module = "%sHandler" % module
        # Look for the component in extensions
        for extension in extensionsByPriority():
            moduleName = [extension, system + "System", cType.capitalize(), module]
            try:
                importedModule = importlib.import_module(moduleName)
                return S_OK(importedModule.__doc__)
            except Exception:
                pass
        return S_ERROR("No documentation was found")

    @staticmethod
    def __storeHostInfo():
        """
        Retrieves and stores into a MySQL database information about the host
        """
        result = SystemAdministratorHandler.__readHostInfo()
        if not result["OK"]:
            gLogger.error(result["Message"])
            return result

        fields = result["Value"]
        fields["Timestamp"] = datetime.utcnow()
        fields["Extension"] = fields["Extensions"]
        client = ComponentMonitoringClient()
        result = client.updateLog(socket.getfqdn(), fields)
        if not result["OK"]:
            gLogger.error(result["Message"])
            return result

        return S_OK("Profiling information logged correctly")

    @staticmethod
    def __deleteOldSoftware(keepLast):
        """
        It removes all versions except the last x

        :param int keepLast: the number of the software version, what we keep
        """
        if six.PY3:
            versionsDirectory = os.path.join(rootPath, "versions")
            versionsDirectoryValid = os.path.isdir(versionsDirectory)
        else:
            versionsDirectory = os.path.split(rootPath)[0]
            # make sure we are not deleting from a wrong directory.
            versionsDirectoryValid = versionsDirectory.endswith("versions")
        if versionsDirectoryValid:
            softwareDirs = {}
            for dirName in os.listdir(versionsDirectory):
                try:
                    # Python 3 uses dashes while Python 2 uses underscores so replace and split
                    # v10.3.1-1637142594, v10r2p10_1629962176
                    version, timestamp = dirName.replace("_", "-").split("-")
                    version = Version(convertToPy3VersionNumber(version))
                    timestamp = int(timestamp)
                except Exception:
                    gLogger.exception("Failed to extract version info from", "%r in %r" % (dirName, versionsDirectory))
                    continue
                softwareDirs[dirName] = (version, timestamp)
            softwareDirs = sorted(softwareDirs, key=softwareDirs.__getitem__, reverse=False)

            try:
                for directoryName in softwareDirs[: -1 * int(keepLast)]:
                    fullPath = os.path.join(versionsDirectory, directoryName)
                    gLogger.info("Removing %s directory." % fullPath)
                    shutil.rmtree(fullPath)
            except Exception as e:
                gLogger.error("Can not delete old DIRAC versions from the file system", repr(e))
        else:
            gLogger.error("The DIRAC.rootPath is not correct:", versionsDirectory)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import datetime
from DIRAC.Core.Utilities.JEncode import JSerializable


class FTS3File(JSerializable):
    """This class represents an a File on which a given Operation
    (Transfer, Staging) should be executed
    """

    # START states
    ALL_STATES = [
        "New",  # Nothing was attempted yet on this file
        "Submitted",  # From FTS: Initial state of a file as soon it's dropped into the database
        "Ready",  # From FTS: File is ready to become active
        "Active",  # From FTS: File went active
        "Finished",  # From FTS: File finished gracefully
        "Canceled",  # From FTS: Canceled by the user
        "Staging",  # From FTS: When staging of a file is requested
        "Failed",  # From FTS: File failure
        "Defunct",  # Totally fail, no more attempt will be made
        "Started",  # From FTS: File transfer has started
        "Not_used",  # From FTS: Transfer not being considered yet, waiting for another one (multihop)
        "Archiving",  # From FTS: file not yet migrated to tape
    ]

    # These are the states that we consider final.
    # No new attempts will be done on our side for
    # FTS3Files reaching one of these
    # Note that Canceled is not final state, because
    # FTS fails some transfers as "Canceled" (gsiftp timeout)

    FINAL_STATES = ["Finished", "Defunct"]

    FTS_SUCCESS_STATES = ["Finished"]
    FTS_FAILED_STATES = ["Canceled", "Failed"]

    # These are the states that the fts servers consider final.
    # No new attempts will be done on their side, but we can
    # still retry.
    FTS_FINAL_STATES = FTS_SUCCESS_STATES + FTS_FAILED_STATES
    INIT_STATE = "New"

    # END states

    _attrToSerialize = [
        "fileID",
        "operationID",
        "status",
        "attempt",
        "creationTime",
        "lastUpdate",
        "rmsFileID",
        "checksum",
        "size",
        "lfn",
        "error",
        "targetSE",
        "ftsGUID",
    ]

    def __init__(self):

        self.status = FTS3File.INIT_STATE
        self.attempt = 0

        now = datetime.datetime.utcnow().replace(microsecond=0)

        self.creationTime = now
        self.lastUpdate = now

        self.rmsFileID = 0
        self.checksum = None
        self.size = 0

        self.lfn = None
        self.error = None

        self.targetSE = None

        # Place holder for the latest job taking care of this file
        self.ftsGUID = None

    @staticmethod
    def fromRMSFile(rmsFile, targetSE):
        """Returns an FTS3File constructed from an RMS File.
        It takes the value of LFN, rmsFileID, checksum and Size

        :param rmsFile: the RMS File to use as source
        :param targetSE: the SE target

        :returns: an FTS3File instance
        """
        ftsFile = FTS3File()
        ftsFile.lfn = rmsFile.LFN
        ftsFile.rmsFileID = rmsFile.FileID
        ftsFile.checksum = rmsFile.Checksum
        ftsFile.size = rmsFile.Size
        ftsFile.targetSE = targetSE

        return ftsFile



#!/usr/bin/env python
########################################################################
# File :    dirac-wms-job-get-jdl
# Author :  Stuart Paterson
########################################################################
"""
Retrieve the current JDL of a DIRAC job

Usage:
  dirac-wms-job-get-jdl [options] ... JobID ...

Arguments:
  JobID:    DIRAC Job ID

Example:
  $ dirac-wms-job-get-jdl 1
  {'Arguments': '-ltrA',
   'CPUTime': '86400',
   'DIRACSetup': 'EELA-Production',
   'Executable': '/bin/ls',
   'JobID': '1',
   'JobName': 'DIRAC_vhamar_602138',
   'JobRequirements': '[OwnerDN = /O=GRID-FR/C=FR/O=CNRS/OU=CPPM/CN=Vanessa Hamar;
                        OwnerGroup = eela_user;
                        Setup = EELA-Production;
                        UserPriority = 1;
                        CPUTime = 0 ]',
   'OutputSandbox': ['std.out', 'std.err'],
   'Owner': 'vhamar',
   'OwnerDN': '/O=GRID-FR/C=FR/O=CNRS/OU=CPPM/CN=Vanessa Hamar',
   'OwnerGroup': 'eela_user',
   'OwnerName': 'vhamar',
   'Priority': '1'}
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id$"

import DIRAC
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


@Script()
def main():
    original = False
    Script.registerSwitch("O", "Original", "Gets the original JDL")
    # Registering arguments will automatically add their description to the help menu
    Script.registerArgument(["JobID:    DIRAC Job ID"])
    sws, args = Script.parseCommandLine(ignoreErrors=True)

    for switch in sws:
        if switch[0] == "Original" or switch[0] == "O":
            original = True

    from DIRAC.Interfaces.API.Dirac import Dirac, parseArguments

    dirac = Dirac()
    exitCode = 0
    errorList = []

    for job in parseArguments(args):

        result = dirac.getJobJDL(job, original=original, printOutput=True)
        if not result["OK"]:
            errorList.append((job, result["Message"]))
            exitCode = 2

    for error in errorList:
        print("ERROR %s: %s" % error)

    DIRAC.exit(exitCode)


if __name__ == "__main__":
    main()



""" This is the guy that parses and interprets the local configuration options.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import re
import os
import sys
import six
import getopt

import DIRAC
from DIRAC import gLogger
from DIRAC import S_OK, S_ERROR

from DIRAC.ConfigurationSystem.Client.ConfigurationData import gConfigurationData
from DIRAC.ConfigurationSystem.private.Refresher import gRefresher
from DIRAC.ConfigurationSystem.Client.PathFinder import getServiceSection, getAgentSection, getExecutorSection
from DIRAC.Core.Utilities.Devloader import Devloader


class LocalConfiguration(object):
    """
    Main class to interface with Configuration of a running DIRAC Component.

    For most cases this is handled via
      - DIRAC.Core.Base.Script class for scripts
      - dirac-agent for agents
      - dirac-service for services
    """

    def __init__(self, defaultSectionPath=""):
        self.currentSectionPath = defaultSectionPath
        self.mandatoryEntryList = []
        self.optionalEntryList = []
        self.commandOptionList = []
        self.commandArgumentList = []
        self.unprocessedSwitches = []
        self.additionalCFGFiles = []
        self.parsedOptionList = []
        self.commandArgList = []
        self.commandGroupArgList = []
        self.cliAdditionalCFGFiles = []
        self.__registerBasicOptions()
        self.isParsed = False
        self.componentName = "Unknown"
        self.componentType = False
        self.loggingSection = "/DIRAC"
        self.initialized = False
        self.__scriptDescription = ""
        self.__helpUsageDoc = ""
        self.__helpArgumentsDoc = ""
        self.__helpExampleDoc = ""
        self.__debugMode = 0
        self.firstOptionIndex = 1

    def disableParsingCommandLine(self):
        self.isParsed = True

    def __getAbsolutePath(self, optionPath):
        if optionPath[0] == "/":
            return optionPath
        else:
            return "%s/%s" % (self.currentSectionPath, optionPath)

    def addMandatoryEntry(self, optionPath):
        """
        Define a mandatory Configuration data option for the parsing of the command line
        """
        self.mandatoryEntryList.append(optionPath)

    def addDefaultEntry(self, optionPath, value):
        """
        Define a default value for a Configuration data option
        """
        if optionPath[0] == "/":
            if not gConfigurationData.extractOptionFromCFG(optionPath):
                self.__setOptionValue(optionPath, value)
        else:
            self.optionalEntryList.append((optionPath, str(value)))

    def addCFGFile(self, filePath):
        """
        Load additional .cfg file to be parsed
        """
        self.additionalCFGFiles.append(filePath)

    def setUsageMessage(self, usageMsg):
        """Define and parse message to be display by the showHelp method.

        :param str usageMsg: script description that can contain Usage, Example, Arguments, Options blocks
        """
        # Searched text
        context = r"(.*?)"
        # Start of any block description or end of __doc__
        startAnyBlockOrEnd = r"(?:\n(?:Usage|Example|Arguments|Options|General options):+\n|$)"

        r = r"%s%s" % (context, startAnyBlockOrEnd)
        if usageMsg:
            # The description block is the first in the __doc__
            desc = re.search(r, usageMsg, re.DOTALL)
            if desc:
                self.__scriptDescription = "\n" + desc.group(1).strip("\n") + "\n"
            # The usage block starts with '\nUsage:\n' or '\nUsage::\n'
            usage = re.search(r"%s%s" % (r"Usage:+", r), usageMsg, re.DOTALL)
            if usage:
                self.__helpUsageDoc = "\nUsage:\n" + usage.group(1).strip("\n")
            # The argument block starts with '\Arguments:\n' or '\Arguments::\n'
            args = re.search(r"%s%s" % (r"Arguments:+", r), usageMsg, re.DOTALL)
            if args:
                self.__helpArgumentsDoc = "\nArguments:\n" + args.group(1).strip("\n")
            # The example block starts with '\Example:\n' or '\Example::\n'
            expl = re.search(r"%s%s" % (r"Example:+", r), usageMsg, re.DOTALL)
            if expl:
                self.__helpExampleDoc = "\nExample:\n" + expl.group(1).strip("\n")

    def __setOptionValue(self, optionPath, value):
        gConfigurationData.setOptionInCFG(self.__getAbsolutePath(optionPath), str(value))

    def __registerBasicOptions(self):
        """Add the basic options::

        General options:
          -o  --option <value>         : Option=value to add
          -s  --section <value>        : Set base section for relative parsed options
          -c  --cert <value>           : Use server certificate to connect to Core Services
          -d  --debug                  : Set debug mode (-ddd is extra debug)
          -   --cfg=                   : Load additional config file
          -   --autoreload             : Automatically restart if there's any change in the module
          -   --license                : Show DIRAC's LICENSE
          -h  --help                   : Shows this help
        """
        self.registerCmdOpt("o:", "option=", "Option=value to add", self.__setOptionByCmd)
        self.registerCmdOpt("s:", "section=", "Set base section for relative parsed options", self.__setSectionByCmd)
        self.registerCmdOpt("c:", "cert=", "Use server certificate to connect to Core Services", self.__setUseCertByCmd)
        self.registerCmdOpt("d", "debug", "Set debug mode (-ddd is extra debug)", self.__setDebugMode)
        self.registerCmdOpt("", "cfg=", "Load additional config file", None)
        devLoader = Devloader()
        if devLoader.enabled:
            self.registerCmdOpt(
                "", "autoreload", "Automatically restart if there's any change in the module", self.__setAutoreload
            )
        self.registerCmdOpt("", "license", "Show DIRAC's LICENSE", self.showLicense)
        self.registerCmdOpt("h", "help", "Shows this help", self.showHelp)

    def registerCmdOpt(self, shortOption, longOption, helpString, function=False):
        """Register a new command line option. The options must be unique and do not coincide with the
        :func:`basic options <__registerBasicOptions>`.

        :param str shortOption: short option name. If the command accepts only long options, then should be ("").
            If such a character is followed by a ':', the option requires an argument.
        :param str longOption: long option name.
            If such a character is followed by a '=', the option requires an argument.
        :param str helpString: option description
        :param function: the callback function to be called with the passed option argument.
            if False, the argument can be retrieved using :func:`getUnprocessedSwitches`.
        """
        shortOption = shortOption.strip()
        longOption = longOption.strip()
        if not shortOption and not longOption:
            raise Exception("No short or long options defined")
        for optTuple in self.commandOptionList:
            if shortOption and optTuple[0] == shortOption:
                raise Exception("Short switch %s is already defined!" % shortOption)
            if longOption and optTuple[1] == longOption:
                raise Exception("Long switch %s is already defined!" % longOption)
        self.commandOptionList.append((shortOption, longOption, helpString, function))

    def registerCmdArg(self, description, mandatory=True, values=None, default=None):
        """Register a new command line argument

        Example::

          # String description type describe simple argument, e.g.:
          registerCmdArg('SingleArg: my single argument')

          # Tuple description type describe argument,
          # which may be replaced by one of those described, e.g.:
          registerCmdArg(('ThisArg: my this argument', 'ThatArg: my that argument'))

          # List description type describe an argument that may have more than one value
          registerCmdArg(['ListArg: my single argument'])

        Result::

          Usage:

            command [options] ... SingleArg <ThisArg|ThatArg> ListArg [ListArg]

          Arguments:

            SingleArg: my single argument
            ThisArg:   my this argument
            ThatArg:   my that argument
            ListArg:   my list of arguments

        :param description: argument description
        :type description: str, tuple or list
        :param bool mandatory: an argument can be mandatory or optional
        :param list values: list of the accepted values
        :param object default: default value
        """
        # Marking an argument in a usage line
        argMarking = ""
        if not description:
            raise Exception("No argument description defined")

        # Single argument, e.g.: Name
        if isinstance(description, six.string_types):
            argMarking = description.split(":")[0].strip()
            description = [description]
        # Single argument that can have two names, e.g.: <User|DN>
        elif isinstance(description, tuple):
            argMarking = "<%s>" % "|".join([d.split(":")[0].strip() for d in description])
        # List arguments, e.g.: CE [CE]
        elif isinstance(description, list):
            argMarking = "{0} [{0}]".format(description[0].split(":")[0].strip())
            if default is None:
                default = []
        else:
            raise Exception("Unknown argument description type.")

        # Check the sequence of the mandatory arguments
        if mandatory and self.commandArgumentList and not self.commandArgumentList[-1][2]:
            raise Exception("The mandatory argument cannot go after the optional one.")

        # Add to others
        description = [tuple([i.strip() for i in d.split(":", 1)]) for d in description]
        self.commandArgumentList.append((argMarking, description, mandatory, values, default))

        # If present list arguments
        listArgs = [t for t in self.commandArgumentList if re.match(r"\w+ \[\w+\]", t[0])]
        if len(listArgs) > 1:
            raise Exception("Can't calculate list of arguments when there are two such arguments of type.")

        if listArgs:
            listArgIndex = self.commandArgumentList.index(listArgs[0])
            listArgLast = self.commandArgumentList[-1] == listArgs[0]
            if not listArgLast and not all([t[2] for t in self.commandArgumentList[listArgIndex:]]):
                raise Exception("Can't calculate list of arguments when not all arguments defined as mandatory.")

    def getExtraCLICFGFiles(self):
        """
        Retrieve list of parsed .cfg files
        """
        if not self.isParsed:
            self.__parseCommandLine()
        return self.cliAdditionalCFGFiles

    def getPositionalArguments(self, group=False):
        """Retrieve list of command line positional arguments

        :param bool group: to return grouped arguments as they were registered

        :return: list
        """
        if not self.isParsed:
            self.__parseCommandLine()

        if group:
            if len(self.commandGroupArgList) == 1:
                return self.commandGroupArgList[0]
            return self.commandGroupArgList

        return self.commandArgList

    def getUnprocessedSwitches(self):
        """
        Retrieve list of command line switches without a callback function
        """
        if not self.isParsed:
            self.__parseCommandLine()
        return self.unprocessedSwitches

    def __checkMandatoryOptions(self):
        try:
            isMandatoryMissing = False
            for optionPath in self.mandatoryEntryList:
                optionPath = self.__getAbsolutePath(optionPath)
                if not gConfigurationData.extractOptionFromCFG(optionPath):
                    gLogger.fatal("Missing mandatory local configuration option", optionPath)
                    isMandatoryMissing = True
            if isMandatoryMissing:
                return S_ERROR()
            return S_OK()
        except Exception as e:
            gLogger.exception()
            return S_ERROR(str(e))

    # TODO: Initialize if not previously initialized
    def initialize(self, componentName):
        """
        Make sure DIRAC is properly initialized
        """
        if self.initialized:
            return S_OK()
        self.initialized = True
        # Set that the command line has already been parsed
        self.isParsed = True
        if not self.componentType:
            self.setConfigurationForScript(componentName)
        try:
            retVal = self.__addUserDataToConfiguration()
            self.__initLogger(self.componentName, self.loggingSection)
            if not retVal["OK"]:
                return retVal
            retVal = self.__checkMandatoryOptions()
            if not retVal["OK"]:
                return retVal
        except Exception as e:
            gLogger.exception()
            return S_ERROR(str(e))
        return S_OK()

    def __initLogger(self, componentName, logSection, forceInit=False):
        gLogger.initialize(componentName, logSection, forceInit=forceInit)

        if self.__debugMode == 1:
            gLogger.setLevel("VERBOSE")
        elif self.__debugMode == 2:
            gLogger.setLevel("VERBOSE")
            gLogger.showHeaders(True)
        elif self.__debugMode >= 3:
            gLogger.setLevel("DEBUG")
            gLogger.showHeaders(True)

    def loadUserData(self):
        """
        This is the magic method that reads the command line and processes it
        It is used by the Script Base class and the dirac-service and dirac-agent scripts
        Before being called:
        - any additional switches to be processed
        - mandatory and default configuration configuration options must be defined.

        """
        if self.initialized:
            return S_OK()
        self.initialized = True
        try:
            retVal = self.__addUserDataToConfiguration()

            for optionTuple in self.optionalEntryList:
                optionPath = self.__getAbsolutePath(optionTuple[0])
                if not gConfigurationData.extractOptionFromCFG(optionPath):
                    gConfigurationData.setOptionInCFG(optionPath, optionTuple[1])

            self.__initLogger(self.componentName, self.loggingSection)
            if not retVal["OK"]:
                return retVal

            retVal = self.__checkMandatoryOptions()
            if not retVal["OK"]:
                return retVal

        except Exception as e:
            gLogger.exception()
            return S_ERROR(str(e))
        return S_OK()

    def __parseCommandLine(self):
        gLogger.debug("Parsing command line")
        shortOption = ""
        longOptionList = []
        for optionTuple in self.commandOptionList:
            if shortOption.find(optionTuple[0]) < 0:
                shortOption += "%s" % optionTuple[0]
            else:
                if optionTuple[0]:
                    gLogger.error("Short option -%s has been already defined" % optionTuple[0])
            if not optionTuple[1] in longOptionList:
                longOptionList.append("%s" % optionTuple[1])
            else:
                if optionTuple[1]:
                    gLogger.error("Long option --%s has been already defined" % optionTuple[1])

        try:
            opts, args = getopt.gnu_getopt(sys.argv[self.firstOptionIndex :], shortOption, longOptionList)
        except getopt.GetoptError as x:
            # x = option "-k" not recognized
            # print help information and exit
            gLogger.fatal("Error when parsing command line arguments: %s" % str(x))
            self.showHelp(exitCode=2)

        for opt, val in opts:
            if opt in ("-h", "--help"):
                self.showHelp()
                sys.exit(2)
            if opt == "--cfg":
                self.cliAdditionalCFGFiles.append(os.path.expanduser(val))

        # environment variable to ensure smooth transition
        if os.getenv("DIRAC_NO_CFG", None):
            self.commandArgList = args
        else:
            # to avoid issuing the warning for correctly passed cfg files
            extraCfg = [os.path.expanduser(arg) for arg in args if arg.endswith(".cfg")]
            self.cliAdditionalCFGFiles.extend(extraCfg)
            self.commandArgList = [arg for arg in args if not arg.endswith(".cfg")]
            if extraCfg:
                # use error level to make sure users will always see the Warning
                gLogger.error(
                    """WARNING: Parsing of '.cfg' files as command line arguments is changing!
          Set the environment variable 'export DIRAC_NO_CFG=1' to pass the file as a positional
          argument (this will become the default).
          To modify the local configuration use '--cfg <configfile>' instead."""
                )
                if os.environ.get("DIRAC_DEPRECATED_FAIL", None):
                    raise NotImplementedError("ERROR: using deprecated config file passing option.")

        # Check arguments
        if len(self.commandArgList) < len([t for t in self.commandArgumentList if t[2]]):
            gLogger.fatal("Error when parsing command line arguments: not all mandatory arguments are defined.")
            self.showHelp(exitCode=1)

        groupArgs = []
        step = 0
        for i in range(len(self.commandArgumentList)):
            argMarking, description, mandatory, values, default = self.commandArgumentList[i]

            # Check whether the required arguments are given in the command line
            if len(self.commandArgList) <= (i + step):
                groupArgs.append(default)
                continue

            # Replace the default value with a defined one
            if re.match(r"\w+ \[\w+\]", argMarking):  # When we consider multiple value argument
                # The index of the last such argument
                step = len(self.commandArgList) - len(self.commandArgumentList)
                groupArgs.append(self.commandArgList[i : i + 1 + step])
            else:
                groupArgs.append(self.commandArgList[i + step])

            # Check accepted values
            for cArg in self.commandArgList[i : i + 1 + step]:
                if values and cArg not in values:
                    gLogger.fatal(
                        "Error when parsing command line arguments: "
                        '"%s" does not match the allowed values for %s' % (cArg, argMarking)
                    )
                    self.showHelp(exitCode=1)

        self.commandGroupArgList = groupArgs
        self.parsedOptionList = opts
        self.isParsed = True

    def __loadCFGFiles(self):
        """
        Loads possibly several cfg files, in order:
        1. ~/.dirac.cfg
        2. cfg files pointed by DIRACSYSCONFIG env variable (comma-separated)
        3. cfg files specified in addCFGFile calls
        4. cfg files that come from the command line
        """
        errorsList = []
        if "DIRACSYSCONFIG" in os.environ:
            diracSysConfigFiles = os.environ["DIRACSYSCONFIG"].replace(" ", "").split(",")
            for diracSysConfigFile in reversed(diracSysConfigFiles):
                gConfigurationData.loadFile(diracSysConfigFile)
        gConfigurationData.loadFile(os.path.expanduser("~/.dirac.cfg"))
        for fileName in self.additionalCFGFiles:
            gLogger.debug("Loading file %s" % fileName)
            retVal = gConfigurationData.loadFile(fileName)
            if not retVal["OK"]:
                gLogger.debug("Could not load file %s: %s" % (fileName, retVal["Message"]))
                errorsList.append(retVal["Message"])
        for fileName in self.cliAdditionalCFGFiles:
            gLogger.debug("Loading file %s" % fileName)
            retVal = gConfigurationData.loadFile(fileName)
            if not retVal["OK"]:
                gLogger.debug("Could not load file %s: %s" % (fileName, retVal["Message"]))
                errorsList.append(retVal["Message"])
        return errorsList

    def __addUserDataToConfiguration(self):
        if not self.isParsed:
            self.__parseCommandLine()

        errorsList = self.__loadCFGFiles()

        if gConfigurationData.getServers():
            retVal = self.syncRemoteConfiguration()
            if not retVal["OK"]:
                return retVal
        else:
            gLogger.warn("Running without remote configuration")

        try:
            if self.componentType == "service":
                self.__setDefaultSection(getServiceSection(self.componentName))
            elif self.componentType == "agent":
                self.__setDefaultSection(getAgentSection(self.componentName))
            elif self.componentType == "executor":
                self.__setDefaultSection(getExecutorSection(self.componentName))
            elif self.componentType == "web":
                self.__setDefaultSection("/%s" % self.componentName)
            elif self.componentType == "script":
                if self.componentName and self.componentName[0] == "/":
                    self.__setDefaultSection(self.componentName)
                    self.componentName = self.componentName[1:]
                else:
                    self.__setDefaultSection("/Scripts/%s" % self.componentName)
            else:
                self.__setDefaultSection("/")
        except Exception as e:
            errorsList.append(str(e))

        self.unprocessedSwitches = []

        for optionName, optionValue in self.parsedOptionList:
            optionName = optionName.lstrip("-")
            for shortOpt, longOpt, _, func in self.commandOptionList:
                if optionName == shortOpt.replace(":", "") or optionName == longOpt.replace("=", ""):
                    if func:
                        retVal = func(optionValue)
                        if not isinstance(retVal, dict):
                            errorsList.append("Callback for switch '%s' does not return S_OK or S_ERROR" % optionName)
                        elif not retVal["OK"]:
                            errorsList.append(retVal["Message"])
                    else:
                        self.unprocessedSwitches.append((optionName, optionValue))

        if len(errorsList) > 0:
            return S_ERROR("\n%s" % "\n".join(errorsList))
        return S_OK()

    def disableCS(self):
        """
        Do not contact Configuration Server upon initialization
        """
        gRefresher.disable()

    def enableCS(self):
        """
        Force the connection the Configuration Server

        (And incidentally reinitialize the ObjectLoader and logger)
        """
        res = gRefresher.enable()

        # This is quite ugly but necessary for the logging
        # We force the reinitialization of the ObjectLoader
        # so that it also takes into account the extensions
        # (since the first time it is loaded by the logger BEFORE the full CS init)
        # And then we regenerate all the backend
        if res["OK"]:
            from DIRAC.Core.Utilities.ObjectLoader import ObjectLoader

            objLoader = ObjectLoader()
            objLoader.reloadRootModules()
            self.__initLogger(self.componentName, self.loggingSection, forceInit=True)
        return res

    def isCSEnabled(self):
        """
        Retrieve current status of the connection to Configuration Server
        """
        return gRefresher.isEnabled()

    def syncRemoteConfiguration(self, strict=False):
        """
        Force a Resync with Configuration Server
        Under normal conditions this is triggered by an access to any
        configuration data.
        """
        if self.componentName == "Configuration/Server":
            if gConfigurationData.isMaster():
                gLogger.info("Starting Master Configuration Server")
                gRefresher.disable()
                return S_OK()
        retDict = gRefresher.forceRefresh()
        if not retDict["OK"]:
            gLogger.error("Can't update from any server", retDict["Message"])
            if strict:
                return retDict
        return S_OK()

    def __setDefaultSection(self, sectionPath):
        self.currentSectionPath = sectionPath
        self.loggingSection = self.currentSectionPath

    def setConfigurationForServer(self, serviceName):
        """
        Declare this is a DIRAC service
        """
        self.componentName = serviceName
        self.componentType = "service"

    def setConfigurationForAgent(self, agentName):
        """
        Declare this is a DIRAC agent
        """
        self.componentName = agentName
        self.componentType = "agent"

    def setConfigurationForExecutor(self, executorName):
        """
        Declare this is a DIRAC agent
        """
        self.componentName = executorName
        self.componentType = "executor"

    def setConfigurationForWeb(self, webName):
        """
        Declare this is a DIRAC agent
        """
        self.componentName = webName
        self.componentType = "web"

    def setConfigurationForScript(self, scriptName):
        """
        Declare this is a DIRAC script
        """
        self.componentName = scriptName
        self.componentType = "script"

    def __setSectionByCmd(self, value):
        if value[0] != "/":
            return S_ERROR("%s is not a valid section. It should start with '/'" % value)
        self.currentSectionPath = value
        return S_OK()

    def __setOptionByCmd(self, value):
        valueList = value.split("=")
        if len(valueList) < 2:
            # FIXME: in the method above an exception is raised, check consitency
            return S_ERROR("-o expects a option=value argument.\nFor example %s -o Port=1234" % sys.argv[0])
        self.__setOptionValue(valueList[0], "=".join(valueList[1:]))
        return S_OK()

    def __setUseCertByCmd(self, value):
        useCert = "no"
        if value.lower() in ("y", "yes", "true"):
            useCert = "yes"
        self.__setOptionValue("/DIRAC/Security/UseServerCertificate", useCert)
        return S_OK()

    def __setDebugMode(self, dummy=False):
        self.__debugMode += 1
        return S_OK()

    def __setAutoreload(self, filepath=False):
        devLoader = Devloader()
        devLoader.bootstrap()
        if filepath:
            devLoader.watchFile(filepath)
        gLogger.notice("Devloader started")
        return S_OK()

    def getDebugMode(self):
        return self.__debugMode

    def showLicense(self, dummy=False):
        """
        Print license
        """
        lpath = os.path.join(DIRAC.rootPath, "DIRAC", "LICENSE")
        sys.stdout.write(" - DIRAC is GPLv3 licensed\n\n")
        try:
            with open(lpath) as fd:
                sys.stdout.write(fd.read())
        except IOError:
            sys.stdout.write("Can't find GPLv3 license at %s. Somebody stole it!\n" % lpath)
            sys.stdout.write("Please check out http://www.gnu.org/licenses/gpl-3.0.html for more info\n")
        DIRAC.exit(0)

    def showHelp(self, dummy=False, exitCode=0):
        """Printout help message including a Usage message if defined via setUsageMessage method

        :param str dummy: dummy text
        :param int exitCode: exit code
        """
        # Intro
        if self.__scriptDescription:
            gLogger.notice(self.__scriptDescription)

        # Usage line
        if self.__helpUsageDoc:
            gLogger.notice(self.__helpUsageDoc)
        else:
            gLogger.notice("\nUsage:")
            gLogger.notice(
                "  %s [options] ..." % os.path.basename(sys.argv[0]).split(".")[0].replace("_", "-"),
                " ".join([t[0] for t in self.commandArgumentList]),
            )
            if dummy:
                gLogger.notice(dummy)

        # Describe options
        gLogger.notice("\nGeneral options:")
        iLastOpt = 0
        for iPos, iVal in enumerate(self.commandOptionList):
            optionTuple = iVal
            if optionTuple[0].endswith(":"):
                line = "  -%s --%s : %s" % (
                    optionTuple[0][:-1].ljust(2),
                    (optionTuple[1][:-1] + " <value> ").ljust(22),
                    optionTuple[2],
                )
                gLogger.notice(line)
            else:
                gLogger.notice("  -%s --%s : %s" % (optionTuple[0].ljust(2), optionTuple[1].ljust(22), optionTuple[2]))
            iLastOpt = iPos
            if optionTuple[0] == "h":
                # Last general opt is always help
                break
        if iLastOpt + 1 < len(self.commandOptionList):
            gLogger.notice("\nOptions:")
            for iPos in range(iLastOpt + 1, len(self.commandOptionList)):
                optionTuple = self.commandOptionList[iPos]
                if optionTuple[0].endswith(":"):
                    line = "  -%s --%s : %s" % (
                        optionTuple[0][:-1].ljust(2),
                        (optionTuple[1][:-1] + " <value> ").ljust(22),
                        optionTuple[2],
                    )
                    gLogger.notice(line)
                else:
                    gLogger.notice(
                        "  -%s --%s : %s" % (optionTuple[0].ljust(2), optionTuple[1].ljust(22), optionTuple[2])
                    )

        # Describe positional arguments
        if self.commandArgumentList:
            # Longest name of the argument
            maxArgLen = 0
            allDescriptions = []
            for argMarking, descriptions, mandatory, values, default in self.commandArgumentList:
                for aName, dText in descriptions:
                    maxArgLen = max(len(aName), maxArgLen)
                    if values:
                        dText += " [%s]" % ", ".join(values)
                    if default:
                        dText += " [default: %s]" % default
                    if not mandatory:
                        dText += " (optional)"
                    # Do not duplicate descriptions
                    if (aName, dText) not in allDescriptions:
                        allDescriptions.append((aName, dText))
            # Print Arguments block
            gLogger.notice("\nArguments:")
            for arg, doc in allDescriptions:
                gLogger.notice("  %s:%s  %s" % (arg, " " * (maxArgLen - len(arg)), doc))
        elif self.__helpArgumentsDoc:
            gLogger.notice(self.__helpArgumentsDoc)

        # Describe example
        if self.__helpExampleDoc:
            gLogger.notice(self.__helpExampleDoc)

        gLogger.notice("")
        DIRAC.exit(exitCode)

    def deleteOption(self, optionPath):
        """
        Remove a Configuration Option from the local Configuration
        """
        gConfigurationData.deleteOptionInCFG(optionPath)



""" The Job Cleaning Agent controls removing jobs from the WMS in the end of their life cycle.

    This agent will take care of:
    - removing all jobs that are in status JobStatus.DELETED
    - deleting (sets status=JobStatus.DELETED) user jobs. The deletion of production jobs should be done by
    :mod:`~DIRAC.TransformationSystem.Agent.TransformationCleaningAgent`.

.. literalinclude:: ../ConfigTemplate.cfg
  :start-after: ##BEGIN JobCleaningAgent
  :end-before: ##END
  :dedent: 2
  :caption: JobCleaningAgent options


Cleaning HeartBeatLoggingInfo
-----------------------------

If the HeartBeatLoggingInfo table of the JobDB is too large, the information for finished jobs can be removed
(including for transformation related jobs).
In vanilla DIRAC the HeartBeatLoggingInfo is only used by the StalledJobAgent. For
this purpose the options MaxHBJobsAtOnce and RemoveStatusDelayHB/[Done|Killed|Failed] should be set to values larger
than 0.

"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import os

from DIRAC import S_OK, S_ERROR
from DIRAC.Core.Base.AgentModule import AgentModule
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
from DIRAC.RequestManagementSystem.Client.Request import Request
from DIRAC.RequestManagementSystem.Client.Operation import Operation
from DIRAC.RequestManagementSystem.Client.File import File
from DIRAC.RequestManagementSystem.Client.ReqClient import ReqClient
from DIRAC.WorkloadManagementSystem.Client import JobStatus
from DIRAC.WorkloadManagementSystem.DB.JobDB import JobDB
from DIRAC.WorkloadManagementSystem.Client.SandboxStoreClient import SandboxStoreClient
from DIRAC.WorkloadManagementSystem.Client.JobMonitoringClient import JobMonitoringClient
from DIRAC.WorkloadManagementSystem.Client.WMSClient import WMSClient

import DIRAC.Core.Utilities.Time as Time


class JobCleaningAgent(AgentModule):
    """
    Agent for removing jobs in status "Deleted", and not only
    """

    def __init__(self, *args, **kwargs):
        """c'tor"""
        AgentModule.__init__(self, *args, **kwargs)

        # clients
        self.jobDB = None

        self.maxJobsAtOnce = 500
        self.prodTypes = []
        self.removeStatusDelay = {}
        self.removeStatusDelayHB = {}

    #############################################################################
    def initialize(self):
        """Sets defaults"""

        self.jobDB = JobDB()

        agentTSTypes = self.am_getOption("ProductionTypes", [])
        if agentTSTypes:
            self.prodTypes = agentTSTypes
        else:
            self.prodTypes = Operations().getValue("Transformations/DataProcessing", ["MCSimulation", "Merge"])
        self.log.info("Will exclude the following Production types from cleaning %s" % (", ".join(self.prodTypes)))
        self.maxJobsAtOnce = self.am_getOption("MaxJobsAtOnce", self.maxJobsAtOnce)

        self.removeStatusDelay[JobStatus.DONE] = self.am_getOption("RemoveStatusDelay/Done", 7)
        self.removeStatusDelay[JobStatus.KILLED] = self.am_getOption("RemoveStatusDelay/Killed", 7)
        self.removeStatusDelay[JobStatus.FAILED] = self.am_getOption("RemoveStatusDelay/Failed", 7)
        self.removeStatusDelay["Any"] = self.am_getOption("RemoveStatusDelay/Any", -1)

        self.removeStatusDelayHB[JobStatus.DONE] = self.am_getOption("RemoveStatusDelayHB/Done", -1)
        self.removeStatusDelayHB[JobStatus.KILLED] = self.am_getOption("RemoveStatusDelayHB/Killed", -1)
        self.removeStatusDelayHB[JobStatus.FAILED] = self.am_getOption("RemoveStatusDelayHB/Failed", -1)
        self.maxHBJobsAtOnce = self.am_getOption("MaxHBJobsAtOnce", 0)

        return S_OK()

    def _getAllowedJobTypes(self):
        """Get valid jobTypes"""
        result = self.jobDB.getDistinctJobAttributes("JobType")
        if not result["OK"]:
            return result
        cleanJobTypes = []
        for jobType in result["Value"]:
            if jobType not in self.prodTypes:
                cleanJobTypes.append(jobType)
        self.log.notice("JobTypes to clean %s" % cleanJobTypes)
        return S_OK(cleanJobTypes)

    def execute(self):
        """Remove or delete jobs in various status"""

        # First, fully remove jobs in JobStatus.DELETED state
        result = self.removeDeletedJobs()
        if not result["OK"]:
            self.log.error("Failed to remove jobs with status %s" % JobStatus.DELETED)

        # Second: set the status to JobStatus.DELETED for certain jobs

        # Get all the Job types for which we can set the status to JobStatus.DELETED
        result = self._getAllowedJobTypes()
        if not result["OK"]:
            return result

        # No jobs in the system subject to deletion
        if not result["Value"]:
            return S_OK()

        baseCond = {"JobType": result["Value"]}
        # Delete jobs with final status
        for status in self.removeStatusDelay:
            delay = self.removeStatusDelay[status]
            if delay < 0:
                # Negative delay means don't delete anything...
                continue
            condDict = dict(baseCond)
            if status != "Any":
                condDict["Status"] = status
            delTime = str(Time.dateTime() - delay * Time.day)
            result = self.deleteJobsByStatus(condDict, delTime)
            if not result["OK"]:
                self.log.error("Failed to delete jobs", "with condDict %s" % condDict)

        if self.maxHBJobsAtOnce > 0:
            for status, delay in self.removeStatusDelayHB.items():
                if delay > 0:
                    self.removeHeartBeatLoggingInfo(status, delay)

        return S_OK()

    def removeDeletedJobs(self):
        """Fully remove jobs that are already in status "DELETED", unless there are still requests.

        :returns: S_OK/S_ERROR
        """

        res = self._getJobsList({"Status": JobStatus.DELETED})
        if not res["OK"]:
            return res
        jobList = res["Value"]
        if not jobList:
            self.log.info("No jobs to remove")
            return S_OK()

        self.log.info("Unassigning sandboxes from soon to be deleted jobs", "(%d)" % len(jobList))
        result = SandboxStoreClient(useCertificates=True).unassignJobs(jobList)
        if not result["OK"]:
            self.log.error("Cannot unassign jobs to sandboxes", result["Message"])
            return result

        self.log.info("Attempting to remove deleted jobs", "(%d)" % len(jobList))

        # remove from jobList those that have still Operations to do in RMS
        reqClient = ReqClient()
        res = reqClient.getRequestIDsForJobs(jobList)
        if not res["OK"]:
            return res
        if res["Value"]["Successful"]:
            notFinal = set()
            # Check whether these requests are in a final status
            for job, reqID in res["Value"]["Successful"].items():
                # If not, remove job from list to remove
                if reqClient.getRequestStatus(reqID).get("Value") not in Request.FINAL_STATES:
                    # Keep that job
                    notFinal.add(job)
                else:
                    # Remove the request, if failed, keep the job
                    res1 = reqClient.deleteRequest(reqID)
                    if not res1["OK"]:
                        notFinal.add(job)
            if notFinal:
                self.log.info(
                    "Some jobs won't be removed, as still having Requests not in final status", "(n=%d)" % len(notFinal)
                )
                jobList = list(set(jobList) - notFinal)
        if not jobList:
            return S_OK()

        ownerJobsDict = self._getOwnerJobsDict(jobList)

        fail = False
        for owner, jobsList in ownerJobsDict.items():
            ownerDN = owner.split(";")[0]
            ownerGroup = owner.split(";")[1]
            self.log.verbose("Attempting to remove jobs", "(n=%d) for %s : %s" % (len(jobsList), ownerDN, ownerGroup))
            wmsClient = WMSClient(useCertificates=True, delegatedDN=ownerDN, delegatedGroup=ownerGroup)
            result = wmsClient.removeJob(jobsList)
            if not result["OK"]:
                self.log.error(
                    "Could not remove jobs",
                    "for %s : %s (n=%d) : %s" % (ownerDN, ownerGroup, len(jobsList), result["Message"]),
                )
                fail = True

        if fail:
            return S_ERROR()

        return S_OK()

    def deleteJobsByStatus(self, condDict, delay=False):
        """Sets the job status to "DELETED" for jobs in condDict.

        :param dict condDict: a dict like {'JobType': 'User', 'Status': 'Killed'}
        :param int delay: days of delay
        :returns: S_OK/S_ERROR
        """

        res = self._getJobsList(condDict, delay)
        if not res["OK"]:
            return res
        jobList = res["Value"]
        if not jobList:
            return S_OK()

        self.log.notice("Attempting to delete jobs", "(%d for %s)" % (len(jobList), condDict))

        result = self.deleteJobOversizedSandbox(jobList)  # This might set a request
        if not result["OK"]:
            self.log.error("Cannot schedule removal of oversized sandboxes", result["Message"])
            return result

        failedJobs = result["Value"][JobStatus.FAILED]
        for job in failedJobs:
            jobList.pop(jobList.index(job))
        if not jobList:
            return S_OK()

        ownerJobsDict = self._getOwnerJobsDict(jobList)

        fail = False
        for owner, jobsList in ownerJobsDict.items():
            ownerDN = owner.split(";")[0]
            ownerGroup = owner.split(";")[1]
            self.log.verbose("Attempting to delete jobs", "(n=%d) for %s : %s" % (len(jobsList), ownerDN, ownerGroup))
            wmsClient = WMSClient(useCertificates=True, delegatedDN=ownerDN, delegatedGroup=ownerGroup)
            result = wmsClient.deleteJob(jobsList)
            if not result["OK"]:
                self.log.error(
                    "Could not delete jobs",
                    "for %s : %s (n=%d) : %s" % (ownerDN, ownerGroup, len(jobsList), result["Message"]),
                )
                fail = True

        if fail:
            return S_ERROR()

        return S_OK()

    def _getJobsList(self, condDict, delay=None):
        """Get jobs list according to conditions

        :param dict condDict: a dict like {'JobType': 'User', 'Status': 'Killed'}
        :param int delay: days of delay
        :returns: S_OK with jobsList
        """
        jobIDsS = set()
        delayStr = "and older than %s" % delay if delay else ""
        self.log.info("Get jobs with %s %s" % (str(condDict), delayStr))
        for order in ["JobID:ASC", "JobID:DESC"]:
            result = self.jobDB.selectJobs(condDict, older=delay, orderAttribute=order, limit=self.maxJobsAtOnce)
            if not result["OK"]:
                return result
            jobIDsS = jobIDsS.union({int(jID) for jID in result["Value"]})

        return S_OK(list(jobIDsS))

    def _getOwnerJobsDict(self, jobList):
        """
        :param list jobList: list of int(JobID)

        :returns: a dict with a grouping of them by owner, e.g.{'dn;group': [1, 3, 4], 'dn;group_1': [5], 'dn_1;group': [2]}
        """
        res = self.jobDB.getJobsAttributes(jobList, ["OwnerDN", "OwnerGroup"])
        if not res["OK"]:
            self.log.error("Could not get the jobs attributes", res["Message"])
            return res
        jobsDictAttribs = res["Value"]

        ownerJobsDict = {}
        for jobID, jobDict in jobsDictAttribs.items():
            ownerJobsDict.setdefault(";".join(jobDict.values()), []).append(jobID)
        return ownerJobsDict

    def deleteJobOversizedSandbox(self, jobIDList):
        """
        Deletes the job oversized sandbox files from storage elements.
        Creates a request in RMS if not immediately possible.

        :param list jobIDList: list of job IDs
        :returns: S_OK/S_ERROR
        """

        failed = {}
        successful = {}

        result = JobMonitoringClient().getJobParameters(jobIDList, ["OutputSandboxLFN"])
        if not result["OK"]:
            return result
        osLFNDict = result["Value"]
        if not osLFNDict:
            return S_OK({"Successful": successful, "Failed": failed})
        osLFNDict = dict(osLFN for osLFN in osLFNDict.items() if osLFN[1])

        self.log.verbose("Deleting oversized sandboxes", osLFNDict)
        # Schedule removal of the LFNs now
        for jobID, outputSandboxLFNdict in osLFNDict.items():  # can be an iterator
            lfn = outputSandboxLFNdict["OutputSandboxLFN"]
            result = self.jobDB.getJobAttributes(jobID, ["OwnerDN", "OwnerGroup"])
            if not result["OK"]:
                failed[jobID] = lfn
                continue
            if not result["Value"]:
                failed[jobID] = lfn
                continue

            ownerDN = result["Value"]["OwnerDN"]
            ownerGroup = result["Value"]["OwnerGroup"]
            result = self.__setRemovalRequest(lfn, ownerDN, ownerGroup)
            if not result["OK"]:
                failed[jobID] = lfn
            else:
                successful[jobID] = lfn

        result = {"Successful": successful, "Failed": failed}
        return S_OK(result)

    def __setRemovalRequest(self, lfn, ownerDN, ownerGroup):
        """Set removal request with the given credentials"""
        oRequest = Request()
        oRequest.OwnerDN = ownerDN
        oRequest.OwnerGroup = ownerGroup
        oRequest.RequestName = os.path.basename(lfn).strip() + "_removal_request.xml"
        oRequest.SourceComponent = "JobCleaningAgent"

        removeFile = Operation()
        removeFile.Type = "RemoveFile"

        removedFile = File()
        removedFile.LFN = lfn

        removeFile.addFile(removedFile)
        oRequest.addOperation(removeFile)

        # put the request with the owner certificate to make sure it's still a valid DN
        return ReqClient(useCertificates=True, delegatedDN=ownerDN, delegatedGroup=ownerGroup).putRequest(oRequest)

    def removeHeartBeatLoggingInfo(self, status, delayDays):
        """Remove HeartBeatLoggingInfo for jobs with given status after given number of days.

        :param str status: Job Status
        :param int delayDays: number of days after which information is removed
        :returns: None
        """
        self.log.info("Removing HeartBeatLoggingInfo for Jobs with %s and older than %s day(s)" % (status, delayDays))
        delTime = str(Time.dateTime() - delayDays * Time.day)
        result = self.jobDB.removeInfoFromHeartBeatLogging(status, delTime, self.maxHBJobsAtOnce)
        if not result["OK"]:
            self.log.error("Failed to delete from HeartBeatLoggingInfo", result["Message"])
        else:
            self.log.info("Deleted HeartBeatLogging info")
        return



#! /usr/bin/env python

# FIXME: if it requires a dirac.cfg it is not a unit test and should be moved to tests directory

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division
import unittest
import time
import os
import shutil
import sys
import six

from DIRAC.Core.Base.Script import parseCommandLine, getPositionalArgs

parseCommandLine()

from DIRAC import gLogger

from DIRAC.Resources.Storage.StorageElement import StorageElement
from DIRAC.Core.Utilities.ReturnValues import returnSingleResult
from DIRAC.Core.Utilities.File import getSize

positionalArgs = getPositionalArgs()

__RCSID__ = "$Id$"

if len(positionalArgs) < 3:
    print("Usage: TestStoragePlugIn.py StorageElement <lfnDir> <localFile>")
    sys.exit()
else:
    storageElementToTest = positionalArgs[0]
    lfnDirToTest = positionalArgs[1]
    fileToTest = positionalArgs[2]


class StorageElementTestCase(unittest.TestCase):
    """Base class for the StorageElement test cases"""

    def setUp(self):
        self.numberOfFiles = 1
        self.storageElement = StorageElement(storageElementToTest)
        self.localSourceFile = fileToTest
        self.localFileSize = getSize(self.localSourceFile)
        self.destDirectory = lfnDirToTest
        # destinationDir = returnSingleResult( self.storageElement.getURL( self.destDirectory ) )['Value']
        destinationDir = self.destDirectory
        res = self.storageElement.createDirectory(destinationDir)
        self.assertTrue(res["OK"])

    def tearDown(self):
        # destinationDir = returnSingleResult( self.storageElement.getURL( self.destDirectory ) )['Value']
        res = self.storageElement.removeDirectory(self.destDirectory, recursive=True)
        self.assertTrue(res["OK"])


class GetInfoTestCase(StorageElementTestCase):
    def test_dump(self):
        print("\n\n#########################################################" "################\n\n\t\t\tDump test\n")
        self.storageElement.dump()

    def test_isValid(self):
        print(
            "\n\n#########################################################" "################\n\n\t\t\tIs valid test\n"
        )
        res = self.storageElement.isValid()
        self.assertTrue(res["OK"])

    def test_getRemotePlugins(self):
        print(
            "\n\n#########################################################"
            "################\n\n\t\t\tGet remote protocols test\n"
        )
        res = self.storageElement.getRemotePlugins()
        self.assertTrue(res["OK"])
        self.assertEqual(type(res["Value"]), list)

    def test_getLocalPlugins(self):
        print(
            "\n\n#########################################################"
            "################\n\n\t\t\tGet local protocols test\n"
        )
        res = self.storageElement.getLocalPlugins()
        self.assertTrue(res["OK"])
        self.assertEqual(type(res["Value"]), list)

    def test_getPlugins(self):
        print(
            "\n\n#########################################################"
            "################\n\n\t\t\tGet protocols test\n"
        )
        res = self.storageElement.getPlugins()
        self.assertTrue(res["OK"])
        self.assertEqual(type(res["Value"]), list)

    # def test_isLocalSE( self ):
    #  print '\n\n#########################################################################\n\n\t\t\tIs local SE test\n'
    #  res = self.storageElement.isLocalSE()
    #  self.assertTrue(res['OK'])
    #  self.assertFalse( res['Value'] )

    # def test_getStorageElementOption( self ):
    #  print '\n\n########################################################################
    #  \n\n\t\t\tGet storage element option test\n'
    #  res = self.storageElement.getStorageElementOption( 'BackendType' )
    #  self.assertTrue(res['OK'])
    #  self.assertEqual( res['Value'], 'DISET' )

    def test_getStorageParameters(self):
        print(
            "\n\n#########################################################"
            "################\n\n\t\t\tGet storage parameters test\n"
        )
        result = self.storageElement.getStorageParameters("DIP")
        self.assertTrue(result["OK"])
        resDict = result["Value"]
        self.assertEqual(resDict["Protocol"], "dips")
        # self.assertEqual( resDict['SpaceToken'], 'LHCb_RAW' )
        # self.assertEqual( resDict['WSUrl'], '/srm/managerv2?SFN=' )
        # self.assertEqual( resDict['Host'], 'srm-lhcb.cern.ch' )
        # self.assertEqual( resDict['Path'], '/castor/cern.ch/grid' )
        # self.assertEqual( resDict['ProtocolName'], 'SRM2' )
        # self.assertEqual( resDict['Port'], '8443' )


class FileTestCases(StorageElementTestCase):
    def test_exists(self):
        print("\n\n#########################################################" "################\n\n\t\t\tExists test\n")
        destinationFilePath = "%s/testFile.%s" % (self.destDirectory, time.time())
        # pfnForLfnRes = self.storageElement.getURL( destinationFilePath )
        # destinationPfn = list(pfnForLfnRes['Value']['Successful'].values())[0]
        fileDict = {destinationFilePath: self.localSourceFile}
        putFileRes = returnSingleResult(self.storageElement.putFile(fileDict))
        # File exists
        existsRes = returnSingleResult(self.storageElement.exists(destinationFilePath))
        # Now remove the destination file
        removeFileRes = returnSingleResult(self.storageElement.removeFile(destinationFilePath))
        # Check removed file
        missingExistsRes = returnSingleResult(self.storageElement.exists(destinationFilePath))
        # Check directories are handled properly
        destinationDir = os.path.dirname(destinationFilePath)
        directoryExistsRes = returnSingleResult(self.storageElement.exists(destinationDir))

        # Check that the put was done correctly
        self.assertTrue(putFileRes["OK"])
        self.assertTrue(putFileRes["Value"])
        self.assertEqual(putFileRes["Value"], self.localFileSize)
        # Check that we checked the file correctly
        self.assertTrue(existsRes["OK"])
        self.assertTrue(existsRes["Value"])
        # Check that the removal was done correctly
        self.assertTrue(removeFileRes["OK"])
        self.assertTrue(removeFileRes["Value"])
        # Check the exists for non existant file
        self.assertTrue(missingExistsRes["OK"])
        self.assertFalse(missingExistsRes["Value"])
        # Check that directories exist
        self.assertTrue(directoryExistsRes["OK"])
        self.assertTrue(directoryExistsRes["Value"])

    def test_isFile(self):
        print(
            "\n\n#########################################################"
            "################\n\n\t\t\tIs file size test\n"
        )
        destinationFilePath = "%s/testFile.%s" % (self.destDirectory, time.time())
        # pfnForLfnRes = returnSingleResult( self.storageElement.getURL( destinationFilePath ) )
        # destinationPfn = pfnForLfnRes['Value']
        fileDict = {destinationFilePath: self.localSourceFile}
        putFileRes = returnSingleResult(self.storageElement.putFile(fileDict))
        # Is a file
        isFileRes = returnSingleResult(self.storageElement.isFile(destinationFilePath))
        # Now remove the destination file
        removeFileRes = returnSingleResult(self.storageElement.removeFile(destinationFilePath))
        # Get metadata for a removed file
        missingIsFileRes = returnSingleResult(self.storageElement.isFile(destinationFilePath))
        # Check directories are handled properly
        destinationDir = os.path.dirname(destinationFilePath)
        directoryIsFileRes = returnSingleResult(self.storageElement.isFile(destinationDir))

        # Check that the put was done correctly
        self.assertTrue(putFileRes["OK"])
        self.assertTrue(putFileRes["Value"])
        self.assertEqual(putFileRes["Value"], self.localFileSize)
        # Check that we checked the file correctly
        self.assertTrue(isFileRes["OK"])
        self.assertTrue(isFileRes["Value"])
        # Check that the removal was done correctly
        self.assertTrue(removeFileRes["OK"])
        self.assertTrue(removeFileRes["Value"])
        # Check the is file for non existant file
        self.assertFalse(missingIsFileRes["OK"])
        expectedError = "File does not exist"
        self.assertTrue(expectedError in missingIsFileRes["Message"])
        # Check that is file operation with a directory
        self.assertTrue(directoryIsFileRes["OK"])
        self.assertFalse(directoryIsFileRes["Value"])

    def test_putFile(self):
        print(
            "\n\n#########################################################" "################\n\n\t\t\tPut file test\n"
        )
        destinationFilePath = "%s/testFile.%s" % (self.destDirectory, time.time())
        # pfnForLfnRes = returnSingleResult( self.storageElement.getURL( destinationFilePath ) )
        # destinationPfn = pfnForLfnRes['Value']
        fileDict = {destinationFilePath: self.localSourceFile}
        putFileRes = returnSingleResult(self.storageElement.putFile(fileDict))
        # Now remove the destination file
        removeFileRes = returnSingleResult(self.storageElement.removeFile(destinationFilePath))

        # Check that the put was done correctly
        self.assertTrue(putFileRes["OK"])
        self.assertTrue(putFileRes["Value"])
        self.assertEqual(putFileRes["Value"], self.localFileSize)
        # Check that the removal was done correctly
        self.assertTrue(removeFileRes["OK"])
        self.assertTrue(removeFileRes["Value"])

    def test_getFile(self):
        print(
            "\n\n#########################################################" "################\n\n\t\t\tGet file test\n"
        )
        destinationFilePath = "%s/testFile.%s" % (self.destDirectory, time.time())
        # pfnForLfnRes = returnSingleResult( self.storageElement.getURL( destinationFilePath ) )
        # destinationPfn = pfnForLfnRes['Value']
        fileDict = {destinationFilePath: self.localSourceFile}
        putFileRes = returnSingleResult(self.storageElement.putFile(fileDict))
        # Now get a local copy of the file
        getFileRes = returnSingleResult(self.storageElement.getFile(destinationFilePath))
        # Now remove the destination file
        removeFileRes = returnSingleResult(self.storageElement.removeFile(destinationFilePath))
        # Clean up the local mess
        os.remove(os.path.basename(destinationFilePath))

        # Check that the put was done correctly
        self.assertTrue(putFileRes["OK"])
        self.assertTrue(putFileRes["Value"])
        self.assertEqual(putFileRes["Value"], self.localFileSize)
        # Check that we got the file correctly
        self.assertTrue(getFileRes["OK"])
        self.assertEqual(getFileRes["Value"], self.localFileSize)
        # Check that the removal was done correctly
        self.assertTrue(removeFileRes["OK"])
        self.assertTrue(removeFileRes["Value"])

    def test_getFileMetadata(self):
        print(
            "\n\n#########################################################"
            "################\n\n\t\t\tGet file metadata test\n"
        )
        destinationFilePath = "%s/testFile.%s" % (self.destDirectory, time.time())
        # pfnForLfnRes = returnSingleResult( self.storageElement.getURL( destinationFilePath ) )
        # destinationPfn = pfnForLfnRes['Value']
        fileDict = {destinationFilePath: self.localSourceFile}
        putFileRes = returnSingleResult(self.storageElement.putFile(fileDict))
        # Get the file metadata
        getFileMetadataRes = returnSingleResult(self.storageElement.getFileMetadata(destinationFilePath))
        # Now remove the destination file
        removeFileRes = returnSingleResult(self.storageElement.removeFile(destinationFilePath))
        # Get metadata for a removed file
        getMissingFileMetadataRes = returnSingleResult(self.storageElement.getFileMetadata(destinationFilePath))
        # Check directories are handled properly
        destinationDir = os.path.dirname(destinationFilePath)
        directoryMetadataRes = returnSingleResult(self.storageElement.getFileMetadata(destinationDir))

        # Check that the put was done correctly
        self.assertTrue(putFileRes["OK"])
        self.assertTrue(putFileRes["Value"])
        self.assertEqual(putFileRes["Value"], self.localFileSize)
        # Check that the metadata was done correctly
        self.assertTrue(getFileMetadataRes["OK"])
        metadataDict = getFileMetadataRes["Value"]

        # Works only for SRM2 plugin
        # self.assertTrue( metadataDict['Cached'] )
        # self.assertFalse( metadataDict['Migrated'] )
        self.assertEqual(metadataDict["Size"], self.localFileSize)
        # Check that the removal was done correctly
        self.assertTrue(removeFileRes["OK"])
        self.assertTrue(removeFileRes["Value"])
        # Check the get metadata for non existant file
        self.assertFalse(getMissingFileMetadataRes["OK"])
        expectedError = "File does not exist"
        self.assertTrue(expectedError in getMissingFileMetadataRes["Message"])
        # Check that metadata operation with a directory
        self.assertFalse(directoryMetadataRes["OK"])
        expectedError = "Supplied path is not a file"
        self.assertTrue(expectedError in directoryMetadataRes["Message"])

    def test_getFileSize(self):
        print(
            "\n\n#########################################################"
            "################\n\n\t\t\tGet file size test\n"
        )
        destinationFilePath = "%s/testFile.%s" % (self.destDirectory, time.time())
        # pfnForLfnRes = returnSingleResult( self.storageElement.getURL( destinationFilePath ) )
        # destinationPfn = pfnForLfnRes['Value']
        fileDict = {destinationFilePath: self.localSourceFile}
        putFileRes = returnSingleResult(self.storageElement.putFile(fileDict))
        # Get the file metadata
        getFileSizeRes = returnSingleResult(self.storageElement.getFileSize(destinationFilePath))
        # Now remove the destination file
        removeFileRes = returnSingleResult(self.storageElement.removeFile(destinationFilePath))
        # Get metadata for a removed file
        getMissingFileSizeRes = returnSingleResult(self.storageElement.getFileSize(destinationFilePath))
        # Check directories are handled properly
        destinationDir = os.path.dirname(destinationFilePath)
        directorySizeRes = returnSingleResult(self.storageElement.getFileSize(destinationDir))

        # Check that the put was done correctly
        self.assertTrue(putFileRes["OK"])
        self.assertTrue(putFileRes["Value"])
        self.assertEqual(putFileRes["Value"], self.localFileSize)
        # Check that the metadata was done correctly
        self.assertTrue(getFileSizeRes["OK"])
        self.assertEqual(getFileSizeRes["Value"], self.localFileSize)
        # Check that the removal was done correctly
        self.assertTrue(removeFileRes["OK"])
        self.assertTrue(removeFileRes["Value"])
        # Check the get metadata for non existant file
        self.assertFalse(getMissingFileSizeRes["OK"])
        expectedError = "File does not exist"
        self.assertTrue(expectedError in getMissingFileSizeRes["Message"])
        # Check that metadata operation with a directory
        self.assertFalse(directorySizeRes["OK"])
        expectedError = "Supplied path is not a file"
        self.assertTrue(expectedError in directorySizeRes["Message"])

    def test_getURL(self):

        print(
            "\n\n#########################################################"
            "################\n\n\t\tGet access url test\n"
        )
        destinationFilePath = "%s/testFile.%s" % (self.destDirectory, time.time())
        # pfnForLfnRes = returnSingleResult( self.storageElement.getURL( destinationFilePath ) )
        # destinationPfn = pfnForLfnRes['Value']
        fileDict = {destinationFilePath: self.localSourceFile}
        putFileRes = returnSingleResult(self.storageElement.putFile(fileDict))
        # Get a transfer url for the file
        getTurlRes = self.storageElement.getURL(destinationFilePath, protocol="dips")
        # Remove the destination file
        removeFileRes = returnSingleResult(self.storageElement.removeFile(destinationFilePath))
        # Get missing turl res
        getMissingTurlRes = self.storageElement.getURL(destinationFilePath, protocol="dips")

        # Check that the put was done correctly
        self.assertTrue(putFileRes["OK"])
        self.assertTrue(putFileRes["Value"])
        self.assertEqual(putFileRes["Value"], self.localFileSize)
        # Check that we can get the tURL properly
        self.assertTrue(getTurlRes["OK"])
        self.assertTrue(getTurlRes["Value"])
        self.assertTrue(isinstance(getTurlRes["Value"], dict))
        self.assertTrue(type(getTurlRes["Value"]["Successful"][destinationFilePath]) in six.string_types)
        # Check that the removal was done correctly
        self.assertTrue(removeFileRes["OK"])
        self.assertTrue(removeFileRes["Value"])

        # Works only for SRM2 plugins
        # # Check that non-existant files are handled correctly
        # self.assertFalse( getMissingTurlRes['OK'] )
        # expectedError = "File does not exist"
        # self.assertTrue( expectedError in getMissingTurlRes['Message'] )


# Works only for SRM2 plugins
#   def test_prestageFile( self ):
#     destinationFilePath = '%s/testFile.%s' % ( self.destDirectory, time.time() )
#     pfnForLfnRes = self.storageElement.getURL( destinationFilePath )
#     destinationPfn = pfnForLfnRes['Value']
#     fileDict = {destinationPfn:self.localSourceFile}
#     putFileRes = self.storageElement.putFile( fileDict, singleFile = True )
#     # Get the file metadata
#     prestageFileRes = self.storageElement.prestageFile( destinationPfn, singleFile = True )
#     # Now remove the destination file
#     removeFileRes = self.storageElement.removeFile( destinationPfn, singleFile = True )
#     # Get metadata for a removed file
#     missingPrestageFileRes = self.storageElement.prestageFile( destinationPfn, singleFile = True )
#
#     # Check that the put was done correctly
#     self.assertTrue( putFileRes['OK'] )
#     self.assertTrue( putFileRes['Value'] )
#     self.assertEqual( putFileRes['Value'], self.localFileSize )
#     # Check that the prestage was done correctly
#     self.assertTrue( prestageFileRes['OK'] )
#     self.assertEqual( type( prestageFileRes['Value'] ), types.StringType )
#     # Check that the removal was done correctly
#     self.assertTrue( removeFileRes['OK'] )
#     self.assertTrue( removeFileRes['Value'] )
#     # Check the prestage for non existant file
#     self.assertFalse( missingPrestageFileRes['OK'] )
#     expectedError = "No such file or directory"
#     self.assertTrue( expectedError in missingPrestageFileRes['Message'] )

# Works only for SRM2 plugins
#   def test_prestageStatus( self ):
#     destinationFilePath = '%s/testFile.%s' % ( self.destDirectory, time.time() )
#     pfnForLfnRes = self.storageElement.getURL( destinationFilePath )
#     destinationPfn = pfnForLfnRes['Value']
#     fileDict = {destinationPfn:self.localSourceFile}
#     putFileRes = self.storageElement.putFile( fileDict, singleFile = True )
#     # Get the file metadata
#     prestageFileRes = self.storageElement.prestageFile( destinationPfn, singleFile = True )
#     srmID = ''
#     if prestageFileRes['OK']:
#       srmID = prestageFileRes['Value']
#     # Take a quick break to allow the SRM to realise the file is available
#     sleepTime = 10
#     print 'Sleeping for %s seconds' % sleepTime
#     time.sleep( sleepTime )
#     # Check that we can monitor the stage request
#     prestageStatusRes = self.storageElement.prestageFileStatus( {destinationPfn:srmID}, singleFile = True )
#     # Now remove the destination file
#     removeFileRes = self.storageElement.removeFile( destinationPfn, singleFile = True )
#
#     # Check that the put was done correctly
#     self.assertTrue( putFileRes['OK'] )
#     self.assertTrue( putFileRes['Value'] )
#     self.assertEqual( putFileRes['Value'], self.localFileSize )
#     # Check that the prestage was done correctly
#     self.assertTrue( prestageFileRes['OK'] )
#     self.assertEqual( type( prestageFileRes['Value'] ), types.StringType )
#     # Check the file was found to be staged
#     self.assertTrue( prestageStatusRes['OK'] )
#     self.assertTrue( prestageStatusRes['Value'] )
#     # Check that the removal was done correctly
#     self.assertTrue( removeFileRes['OK'] )
#     self.assertTrue( removeFileRes['Value'] )


# Works only for SRM2 plugins
#   def test_pinRelease( self ):
#     print '\n\n#########################################################################\n\n\t\tPin release test\n'
#     destinationFilePath = '%s/testFile.%s' % ( self.destDirectory, time.time() )
#     pfnForLfnRes = self.storageElement.getURL( destinationFilePath )
#     destinationPfn = pfnForLfnRes['Value']
#     fileDict = {destinationPfn:self.localSourceFile}
#     putFileRes = self.storageElement.putFile( fileDict, singleFile = True )
#     # Get the file metadata
#     pinFileRes = self.storageElement.pinFile( destinationPfn, singleFile = True )
#     srmID = ''
#     if pinFileRes['OK']:
#       srmID = pinFileRes['Value']
#     # Check that we can release the file
#     releaseFileRes = self.storageElement.releaseFile( {destinationPfn:srmID}, singleFile = True )
#     # Now remove the destination file
#     removeFileRes = self.storageElement.removeFile( destinationPfn, singleFile = True )
#
#     # Check that the put was done correctly
#     self.assertTrue( putFileRes['OK'] )
#     self.assertTrue( putFileRes['Value'] )
#     self.assertEqual( putFileRes['Value'], self.localFileSize )
#     # Check that the file pin was done correctly
#     self.assertTrue( pinFileRes['OK'] )
#     self.assertEqual( type( pinFileRes['Value'] ), types.StringType )
#     # Check the file was found to be staged
#     self.assertTrue( releaseFileRes['OK'] )
#     self.assertTrue( releaseFileRes['Value'] )
#     # Check that the removal was done correctly
#     self.assertTrue( removeFileRes['OK'] )
#     self.assertTrue( removeFileRes['Value'] )


class DirectoryTestCases(StorageElementTestCase):
    def test_createDirectory(self):
        print(
            "\n\n#########################################################"
            "################\n\n\t\t\tCreate directory test\n"
        )
        directory = "%s/%s" % (self.destDirectory, "createDirectoryTest")
        # pfnForLfnRes = returnSingleResult( self.storageElement.getURL( directory ) )
        # directoryPfn = pfnForLfnRes['Value']

        createDirRes = self.storageElement.createDirectory(directory)
        # Remove the target dir
        removeDirRes = self.storageElement.removeDirectory(directory, recursive=True)

        # Check that the creation was done correctly
        self.assertTrue(createDirRes["OK"])
        self.assertTrue(createDirRes["Value"])
        # Remove the directory
        self.assertTrue(removeDirRes["OK"])
        self.assertTrue(removeDirRes["Value"])

    def test_isDirectory(self):
        print(
            "\n\n#########################################################"
            "################\n\n\t\t\tIs directory test\n"
        )
        destDirectory = self.destDirectory
        # Test that it is a directory
        isDirectoryRes = self.storageElement.isDirectory(destDirectory)
        # Test that no existant dirs are handled correctly
        nonExistantDir = "%s/%s" % (destDirectory, "NonExistant")
        nonExistantDirRes = self.storageElement.isDirectory(nonExistantDir)

        # Check that it works with the existing dir
        self.assertTrue(isDirectoryRes["OK"])
        self.assertTrue(isDirectoryRes["Value"])
        # Check that we handle non existant correctly
        self.assertTrue(nonExistantDirRes["Value"]["Failed"][nonExistantDir] in ["Path does not exist"])

    def test_listDirectory(self):
        print(
            "\n\n#########################################################"
            "################\n\n\t\t\tList directory test\n"
        )
        destDirectory = "%s/%s" % (self.destDirectory, "listDirectoryTest")
        # destDirectory = returnSingleResult( self.storageElement.getURL( directory ) )['Value']
        # Create a local directory to upload
        localDir = "/tmp/unit-test"
        srcFile = "/etc/group"
        sizeOfLocalFile = getSize(srcFile)
        if not os.path.exists(localDir):
            os.mkdir(localDir)
        for i in range(self.numberOfFiles):
            shutil.copy(srcFile, "%s/testFile.%s" % (localDir, time.time()))
            time.sleep(1)
        # Check that we can successfully upload the directory to the storage element
        dirDict = {destDirectory: localDir}
        putDirRes = self.storageElement.putDirectory(dirDict)
        print(putDirRes)
        # List the remote directory
        listDirRes = self.storageElement.listDirectory(destDirectory)
        # Now remove the remove directory
        removeDirRes = self.storageElement.removeDirectory(destDirectory, recursive=True)
        print(removeDirRes)
        # Clean up the locally created directory
        shutil.rmtree(localDir)

        # Perform the checks for the put dir operation
        self.assertTrue(putDirRes["OK"])
        self.assertTrue(putDirRes["Value"])
        if putDirRes["Value"]["Successful"][destDirectory]["Files"]:
            self.assertEqual(putDirRes["Value"]["Successful"][destDirectory]["Files"], self.numberOfFiles)
            self.assertEqual(
                putDirRes["Value"]["Successful"][destDirectory]["Size"], self.numberOfFiles * sizeOfLocalFile
            )
        self.assertTrue(type(putDirRes["Value"]["Successful"][destDirectory]["Files"]) in six.integer_types)
        self.assertTrue(type(putDirRes["Value"]["Successful"][destDirectory]["Size"]) in six.integer_types)
        # Perform the checks for the list dir operation
        self.assertTrue(listDirRes["OK"])
        self.assertTrue(listDirRes["Value"])
        self.assertTrue("SubDirs" in listDirRes["Value"]["Successful"][destDirectory])
        self.assertTrue("Files" in listDirRes["Value"]["Successful"][destDirectory])
        self.assertEqual(len(listDirRes["Value"]["Successful"][destDirectory]["Files"]), self.numberOfFiles)
        # Perform the checks for the remove directory operation
        self.assertTrue(removeDirRes["OK"])
        self.assertTrue(removeDirRes["Value"])
        if removeDirRes["Value"]["Successful"][destDirectory]["FilesRemoved"]:
            self.assertEqual(removeDirRes["Value"]["Successful"][destDirectory]["FilesRemoved"], self.numberOfFiles)
            self.assertEqual(
                removeDirRes["Value"]["Successful"][destDirectory]["SizeRemoved"], self.numberOfFiles * sizeOfLocalFile
            )
        self.assertTrue(type(removeDirRes["Value"]["Successful"][destDirectory]["FilesRemoved"]) in six.integer_types)
        self.assertTrue(type(removeDirRes["Value"]["Successful"][destDirectory]["SizeRemoved"]) in six.integer_types)

    def test_getDirectoryMetadata(self):
        print(
            "\n\n#########################################################"
            "################\n\n\t\t\tDirectory metadata test\n"
        )
        destDirectory = "%s/%s" % (self.destDirectory, "getDirectoryMetadataTest")
        # destDirectory = returnSingleResult( self.storageElement.getURL( directory ) )['Value']
        # Create a local directory to upload
        localDir = "/tmp/unit-test"
        srcFile = "/etc/group"
        sizeOfLocalFile = getSize(srcFile)
        if not os.path.exists(localDir):
            os.mkdir(localDir)
        for i in range(self.numberOfFiles):
            shutil.copy(srcFile, "%s/testFile.%s" % (localDir, time.time()))
            time.sleep(1)
        # Check that we can successfully upload the directory to the storage element
        dirDict = {destDirectory: localDir}
        putDirRes = self.storageElement.putDirectory(dirDict)
        # Get the directory metadata
        metadataDirRes = self.storageElement.getDirectoryMetadata(destDirectory)
        # Now remove the remove directory
        removeDirRes = self.storageElement.removeDirectory(destDirectory, recursive=True)
        # Clean up the locally created directory
        shutil.rmtree(localDir)

        # Perform the checks for the put dir operation
        self.assertTrue(putDirRes["OK"])
        self.assertTrue(putDirRes["Value"])
        if putDirRes["Value"]["Successful"][destDirectory]["Files"]:
            self.assertEqual(putDirRes["Value"]["Successful"][destDirectory]["Files"], self.numberOfFiles)
            self.assertEqual(
                putDirRes["Value"]["Successful"][destDirectory]["Size"], self.numberOfFiles * sizeOfLocalFile
            )
            self.assertTrue(type(putDirRes["Value"]["Successful"][destDirectory]["Files"]) in six.integer_types)
            self.assertTrue(type(putDirRes["Value"]["Successful"][destDirectory]["Size"]) in six.integer_types)
        # Perform the checks for the list dir operation
        self.assertTrue(metadataDirRes["OK"])
        self.assertTrue(metadataDirRes["Value"])

        # Works only for the SRM2 plugin
        # self.assertTrue( metadataDirRes['Value']['Mode'] )
        # self.assertTrue( type( metadataDirRes['Value']['Mode'] ) == int )

        self.assertTrue(metadataDirRes["Value"]["Successful"][destDirectory]["Exists"])
        self.assertEqual(metadataDirRes["Value"]["Successful"][destDirectory]["Type"], "Directory")
        # Perform the checks for the remove directory operation
        self.assertTrue(removeDirRes["OK"])
        self.assertTrue(removeDirRes["Value"])
        if removeDirRes["Value"]["Successful"][destDirectory]["FilesRemoved"]:
            self.assertEqual(removeDirRes["Value"]["Successful"][destDirectory]["FilesRemoved"], self.numberOfFiles)
            self.assertEqual(
                removeDirRes["Value"]["Successful"][destDirectory]["SizeRemoved"], self.numberOfFiles * sizeOfLocalFile
            )
            self.assertTrue(
                type(removeDirRes["Value"]["Successful"][destDirectory]["FilesRemoved"]) in six.integer_types
            )
            self.assertTrue(
                type(removeDirRes["Value"]["Successful"][destDirectory]["SizeRemoved"]) in six.integer_types
            )

    def test_getDirectorySize(self):
        print(
            "\n\n#########################################################"
            "################\n\n\t\t\tGet directory size test\n"
        )
        destDirectory = "%s/%s" % (self.destDirectory, "getDirectorySizeTest")
        # destDirectory = returnSingleResult( self.storageElement.getURL( directory ) )['Value']
        # Create a local directory to upload
        localDir = "/tmp/unit-test"
        srcFile = "/etc/group"
        sizeOfLocalFile = getSize(srcFile)
        if not os.path.exists(localDir):
            os.mkdir(localDir)
        for i in range(self.numberOfFiles):
            shutil.copy(srcFile, "%s/testFile.%s" % (localDir, time.time()))
            time.sleep(1)
        # Check that we can successfully upload the directory to the storage element
        dirDict = {destDirectory: localDir}
        putDirRes = self.storageElement.putDirectory(dirDict)
        # Get the directory metadata
        getDirSizeRes = self.storageElement.getDirectorySize(destDirectory)
        # Now remove the remove directory
        removeDirRes = self.storageElement.removeDirectory(destDirectory, recursive=True)
        # Clean up the locally created directory
        shutil.rmtree(localDir)

        # Perform the checks for the put dir operation
        self.assertTrue(putDirRes["OK"])
        self.assertTrue(putDirRes["Value"])
        if putDirRes["Value"]["Successful"][destDirectory]["Files"]:
            self.assertEqual(putDirRes["Value"]["Successful"][destDirectory]["Files"], self.numberOfFiles)
            self.assertEqual(
                putDirRes["Value"]["Successful"][destDirectory]["Size"], self.numberOfFiles * sizeOfLocalFile
            )
            self.assertTrue(type(putDirRes["Value"]["Successful"][destDirectory]["Files"]) in six.integer_types)
            self.assertTrue(type(putDirRes["Value"]["Successful"][destDirectory]["Size"]) in six.integer_types)
        # Perform the checks for the get dir size operation
        self.assertTrue(getDirSizeRes["OK"])
        self.assertTrue(getDirSizeRes["Value"])
        self.assertFalse(getDirSizeRes["Value"]["Successful"][destDirectory]["SubDirs"])
        self.assertTrue(type(getDirSizeRes["Value"]["Successful"][destDirectory]["Files"]) in six.integer_types)
        self.assertTrue(type(getDirSizeRes["Value"]["Successful"][destDirectory]["Size"]) in six.integer_types)
        # Perform the checks for the remove directory operation
        self.assertTrue(removeDirRes["OK"])
        self.assertTrue(removeDirRes["Value"])
        if removeDirRes["Value"]["Successful"][destDirectory]["FilesRemoved"]:
            self.assertEqual(removeDirRes["Value"]["Successful"][destDirectory]["FilesRemoved"], self.numberOfFiles)
            self.assertEqual(
                removeDirRes["Value"]["Successful"][destDirectory]["SizeRemoved"], self.numberOfFiles * sizeOfLocalFile
            )
            self.assertTrue(
                type(removeDirRes["Value"]["Successful"][destDirectory]["FilesRemoved"]) in six.integer_types
            )
            self.assertTrue(
                type(removeDirRes["Value"]["Successful"][destDirectory]["SizeRemoved"]) in six.integer_types
            )

    def test_removeDirectory(self):
        print(
            "\n\n#########################################################"
            "################\n\n\t\t\tRemove directory test\n"
        )
        destDirectory = "%s/%s" % (self.destDirectory, "removeDirectoryTest")
        # destDirectory = returnSingleResult( self.storageElement.getURL( directory ) )['Value']
        # Create a local directory to upload
        localDir = "/tmp/unit-test"
        srcFile = "/etc/group"
        sizeOfLocalFile = getSize(srcFile)
        if not os.path.exists(localDir):
            os.mkdir(localDir)
        for i in range(self.numberOfFiles):
            shutil.copy(srcFile, "%s/testFile.%s" % (localDir, time.time()))
            time.sleep(1)
        # Check that we can successfully upload the directory to the storage element
        dirDict = {destDirectory: localDir}
        putDirRes = self.storageElement.putDirectory(dirDict)
        # Get the directory metadata
        # Now remove the remove directory
        removeDirRes = self.storageElement.removeDirectory(destDirectory, recursive=True)
        # Clean up the locally created directory
        shutil.rmtree(localDir)

        # Perform the checks for the put dir operation
        self.assertTrue(putDirRes["OK"])
        self.assertTrue(putDirRes["Value"])
        if putDirRes["Value"]["Successful"][destDirectory]["Files"]:
            self.assertEqual(putDirRes["Value"]["Successful"][destDirectory]["Files"], self.numberOfFiles)
            self.assertEqual(
                putDirRes["Value"]["Successful"][destDirectory]["Size"], self.numberOfFiles * sizeOfLocalFile
            )
            self.assertTrue(type(putDirRes["Value"]["Successful"][destDirectory]["Files"]) in six.integer_types)
            self.assertTrue(type(putDirRes["Value"]["Successful"][destDirectory]["Size"]) in six.integer_types)
        # Perform the checks for the remove directory operation
        self.assertTrue(removeDirRes["OK"])
        self.assertTrue(removeDirRes["Value"])
        if removeDirRes["Value"]["Successful"][destDirectory]["FilesRemoved"]:
            self.assertEqual(removeDirRes["Value"]["Successful"][destDirectory]["FilesRemoved"], self.numberOfFiles)
            self.assertEqual(
                removeDirRes["Value"]["Successful"][destDirectory]["SizeRemoved"], self.numberOfFiles * sizeOfLocalFile
            )
            self.assertTrue(
                type(removeDirRes["Value"]["Successful"][destDirectory]["FilesRemoved"]) in six.integer_types
            )
            self.assertTrue(
                type(removeDirRes["Value"]["Successful"][destDirectory]["SizeRemoved"]) in six.integer_types
            )

    def test_getDirectory(self):
        print(
            "\n\n#########################################################"
            "################\n\n\t\t\tGet directory test\n"
        )
        destDirectory = "%s/%s" % (self.destDirectory, "getDirectoryTest")
        # destDirectory = returnSingleResult( self.storageElement.getURL( directory ) )['Value']
        # Create a local directory to upload
        localDir = "/tmp/unit-test"
        srcFile = "/etc/group"
        sizeOfLocalFile = getSize(srcFile)
        if not os.path.exists(localDir):
            os.mkdir(localDir)
        for i in range(self.numberOfFiles):
            shutil.copy(srcFile, "%s/testFile.%s" % (localDir, time.time()))
            time.sleep(1)
        # Check that we can successfully upload the directory to the storage element
        dirDict = {destDirectory: localDir}
        putDirRes = self.storageElement.putDirectory(dirDict)
        # Get the directory metadata
        # Clean up the locally created directory
        shutil.rmtree(localDir)
        getDirRes = self.storageElement.getDirectory(destDirectory, localPath=localDir)
        # Now remove the remove directory
        removeDirRes = self.storageElement.removeDirectory(destDirectory, recursive=True)
        # Clean up the locally created directory
        if os.path.exists(localDir):
            shutil.rmtree(localDir)

        # Perform the checks for the put dir operation
        self.assertTrue(putDirRes["OK"])
        self.assertTrue(putDirRes["Value"])
        for _dir in dirDict:
            if putDirRes["Value"]["Successful"][_dir]["Files"]:
                self.assertEqual(putDirRes["Value"]["Successful"][_dir]["Files"], self.numberOfFiles)
                self.assertEqual(putDirRes["Value"]["Successful"][_dir]["Size"], self.numberOfFiles * sizeOfLocalFile)
                self.assertTrue(type(putDirRes["Value"]["Successful"][_dir]["Files"]) in six.integer_types)
                self.assertTrue(type(putDirRes["Value"]["Successful"][_dir]["Size"]) in six.integer_types)
        # Perform the checks for the get directory operation
        self.assertTrue(getDirRes["OK"])
        self.assertTrue(getDirRes["Value"])
        for _dir in dirDict:
            if getDirRes["Value"]["Successful"][_dir]["Files"]:
                self.assertEqual(getDirRes["Value"]["Successful"][_dir]["Files"], self.numberOfFiles)
                self.assertEqual(getDirRes["Value"]["Successful"][_dir]["Size"], self.numberOfFiles * sizeOfLocalFile)
                self.assertTrue(type(getDirRes["Value"]["Successful"][_dir]["Files"]) in six.integer_types)
                self.assertTrue(type(getDirRes["Value"]["Successful"][_dir]["Size"]) in six.integer_types)
        # Perform the checks for the remove directory operation
        self.assertTrue(removeDirRes["OK"])
        self.assertTrue(removeDirRes["Value"])
        if removeDirRes["Value"]["Successful"][destDirectory]["FilesRemoved"]:
            self.assertEqual(removeDirRes["Value"]["Successful"][destDirectory]["FilesRemoved"], self.numberOfFiles)
            self.assertEqual(
                removeDirRes["Value"]["Successful"][destDirectory]["SizeRemoved"], self.numberOfFiles * sizeOfLocalFile
            )
            self.assertTrue(
                type(removeDirRes["Value"]["Successful"][destDirectory]["FilesRemoved"]) in six.integer_types
            )
            self.assertTrue(
                type(removeDirRes["Value"]["Successful"][destDirectory]["SizeRemoved"]) in six.integer_types
            )


if __name__ == "__main__":
    gLogger.setLevel("DEBUG")
    suite = unittest.defaultTestLoader.loadTestsFromTestCase(DirectoryTestCases)
    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(FileTestCases))
    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(GetInfoTestCase))
    testResult = unittest.TextTestRunner(verbosity=2).run(suite)
